from copy import deepcopy

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from conftest import notebooks
from xclim.testing.helpers import test_timeseries as timeseries

import xscen as xs


class TestSearchDataCatalogs:
    cat = xs.DataCatalog(notebooks / "samples" / "pangeo-cmip6.json")

    @pytest.mark.parametrize(
        "variables_and_freqs, other_arg",
        [
            ({"tasmin": "D"}, None),
            ({"sftlf": "fx"}, "other"),
            ({"tasmin": "D", "sftlf": "fx"}, "exclusion"),
        ],
    )
    def test_basic(self, variables_and_freqs, other_arg):
        out = xs.search_data_catalogs(
            data_catalogs=self.cat,
            variables_and_freqs=variables_and_freqs,
            other_search_criteria=(
                {"experiment": ["ssp585"]} if other_arg == "other" else None
            ),
            exclusions=(
                {"member": "r2.*", "domain": ["gr2"]}
                if other_arg == "exclusion"
                else None
            ),
        )
        assert len(out) == 13 if other_arg is None else 2 if other_arg == "other" else 4

    @pytest.mark.parametrize(
        "periods, coverage_kwargs, matches",
        [
            ([["2020", "2030"], ["2035", "2040"]], None, 5),
            ([["2020-01", "2030-12"], ["2035-01", "2040-12"]], None, 5),
            ([["2020-02", "2030-11"], ["2035", "2040"]], None, 5),
            ([["1900", "2030"], ["2035", "2040"]], None, 0),
            ([["1985-01-01", "2014-12-31"]], {"coverage": 1}, 8),  # 8
            ([["1985-01-01", "2015-01-01"]], {"coverage": 1}, 0),  # 0
            ([["1985-01-01", "2015-01-01"]], None, 8),  # default coverage = 0.99
            ([["1985-01-01", "2015-04-20"]], None, 8),
            ([["1985-01-01", "2015-04-20 16:00:01"]], None, 0),  # 0
            ([["1985-01-01", "2015-04-20 15:59:00"]], None, 8),  # 8
            ([["1985-01-02", "2015-04-20 15:45:00"]], None, 8),  # 8
            (
                [["1985-01-02", "2015-04-20 15:46:00"]],
                None,
                0,
            ),  # ensure uses intersection
            ([["1985-01-01", "2020"]], None, 0),  # 0
            ([["2020", "2130"]], {"coverage": 0.70}, 5),
        ],
    )
    def test_periods(self, periods, coverage_kwargs, matches):
        out = xs.search_data_catalogs(
            data_catalogs=self.cat,
            variables_and_freqs={"tasmin": "D"},
            periods=periods,
            coverage_kwargs=coverage_kwargs,
        )
        assert len(out) == matches

    def test_ids(self):
        out = xs.search_data_catalogs(
            data_catalogs=deepcopy(self.cat),
            variables_and_freqs={"tasmin": "D"},
            id_columns=["source"],
        )
        assert len(out) == 3
        assert len(out["NorESM2-MM"].df) == 5

    @pytest.mark.parametrize("allow_resampling", [True, False])
    def test_allow_resampling(self, allow_resampling):
        out = xs.search_data_catalogs(
            data_catalogs=deepcopy(self.cat),
            variables_and_freqs={"tasmin": "YS"},
            allow_resampling=allow_resampling,
        )
        assert len(out) == (13 if allow_resampling else 0)

    @pytest.mark.parametrize(
        "restrict_warming_level,exp",
        [(True, 5), ({"wl": 2, "ignore_member": True}, 5), ({"wl": 4}, 2)],
    )
    def test_warminglevel(self, restrict_warming_level, exp):
        cat = deepcopy(self.cat)
        new_line = deepcopy(cat.df.iloc[13])
        new_line["experiment"] = "ssp245"
        new_line["id"] = xs.catalog.generate_id(new_line.to_frame().T).iloc[0]
        cat.esmcat._df = pd.concat([cat.df, new_line.to_frame().T], ignore_index=True)

        out = xs.search_data_catalogs(
            data_catalogs=cat,
            variables_and_freqs={"tasmax": "D"},
            restrict_warming_level=restrict_warming_level,
        )
        assert len(out) == exp

    @pytest.mark.parametrize("restrict_resolution", [None, "finest", "coarsest"])
    def test_restrict_resolution(self, restrict_resolution):
        cat = deepcopy(self.cat)
        for i in range(2):
            new_line = deepcopy(cat.df.iloc[0])
            new_line["mip_era"] = "CMIP5"
            new_line["activity"] = "CORDEX"
            new_line["institution"] = "CCCma"
            new_line["driving_model"] = "CanESM2"
            new_line["source"] = "CRCM5"
            new_line["experiment"] = "rcp85"
            new_line["member"] = "r1i1p1"
            new_line["domain"] = "NAM-22" if i == 0 else "NAM-11"
            new_line["frequency"] = "day"
            new_line["xrfreq"] = "D"
            new_line["variable"] = ("tasmin",)
            new_line["id"] = xs.catalog.generate_id(new_line.to_frame().T).iloc[0]

            cat.esmcat._df = pd.concat(
                [cat.df, new_line.to_frame().T], ignore_index=True
            )

        out = xs.search_data_catalogs(
            data_catalogs=cat,
            variables_and_freqs={"tasmin": "D"},
            other_search_criteria={
                "source": ["GFDL-CM4", "CRCM5"],
                "experiment": ["ssp585", "rcp85"],
            },
            restrict_resolution=restrict_resolution,
        )
        if restrict_resolution is None:
            assert len(out) == 4
        elif restrict_resolution == "finest":
            assert len(out) == 2
            assert any("NAM-11" in x for x in out)
            assert any("_gr1" in x for x in out)
        elif restrict_resolution == "coarsest":
            assert len(out) == 2
            assert any("NAM-22" in x for x in out)
            assert any("_gr2" in x for x in out)

    @pytest.mark.parametrize("restrict_members", [None, {"ordered": 2}])
    def test_restrict_members(self, restrict_members):
        out = xs.search_data_catalogs(
            data_catalogs=self.cat,
            variables_and_freqs={"tasmin": "D"},
            other_search_criteria={
                "source": ["NorESM2-LM"],
                "experiment": ["historical"],
            },
            restrict_members=restrict_members,
        )
        assert len(out) == (3 if restrict_members is None else 2)
        if restrict_members is not None:
            assert all(
                o in out.keys()
                for o in [
                    "CMIP_NCC_NorESM2-LM_historical_r1i1p1f1_gn",
                    "CMIP_NCC_NorESM2-LM_historical_r2i1p1f1_gn",
                ]
            )

        # Make sure that those with fewer members are still returned
        assert (
            len(
                xs.search_data_catalogs(
                    data_catalogs=self.cat,
                    variables_and_freqs={"tasmin": "D"},
                    other_search_criteria={
                        "source": ["GFDL-CM4"],
                        "experiment": ["ssp585"],
                        "domain": "gr1",
                    },
                    restrict_members=restrict_members,
                )
            )
            == 1
        )

    @pytest.mark.parametrize("allow_conversion", [True, False])
    def test_allow_conversion(self, allow_conversion):
        out = xs.search_data_catalogs(
            data_catalogs=self.cat,
            variables_and_freqs={"evspsblpot": "D"},
            other_search_criteria={
                "institution": ["NOAA-GFDL"],
                "experiment": ["ssp585"],
            },
            allow_conversion=allow_conversion,
        )
        assert len(out) == (2 if allow_conversion else 0)
        if allow_conversion:
            assert all(
                v in out[list(out.keys())[0]].unique("variable")
                for v in ["tasmin", "tasmax"]
            )
            assert "tas" not in out[list(out.keys())[0]].unique("variable")

    def test_no_match(self):
        out = xs.search_data_catalogs(
            data_catalogs=self.cat,
            variables_and_freqs={"tas": "YS"},
            allow_resampling=False,
        )
        assert isinstance(out, dict)
        assert len(out) == 0
        out = xs.search_data_catalogs(
            data_catalogs=self.cat,
            variables_and_freqs={"tas": "D"},
            other_search_criteria={"experiment": "not_real"},
        )
        assert isinstance(out, dict)
        assert len(out) == 0

    def test_input_types(self, samplecat):
        data_catalogs_2 = notebooks / "samples" / "pangeo-cmip6.json"

        assert (
            xs.search_data_catalogs(
                data_catalogs=[samplecat, data_catalogs_2],
                variables_and_freqs={"tas": "D"},
                other_search_criteria={
                    "experiment": "ssp585",
                    "source": "NorESM.*",
                    "member": "r1i1p1f1",
                },
            ).keys()
            == xs.search_data_catalogs(
                data_catalogs=[samplecat, self.cat],
                variables_and_freqs={"tas": "D"},
                other_search_criteria={
                    "experiment": "ssp585",
                    "source": "NorESM.*",
                    "member": "r1i1p1f1",
                },
            ).keys()
        )

    def test_match_histfut(self):
        out = xs.search_data_catalogs(
            data_catalogs=self.cat,
            variables_and_freqs={"tasmin": "D"},
            other_search_criteria={"experiment": "ssp585", "source": "GFDL-CM4"},
            match_hist_and_fut=True,
        )
        k = list(out.keys())[0]
        assert str(sorted(out[k].unique("date_start"))[0]) == "1985-01-01 00:00:00"
        assert str(sorted(out[k].unique("date_start"))[1]) == "2015-01-01 00:00:00"

    def test_fx(self):
        cat = deepcopy(self.cat)
        new_line = deepcopy(cat.df.iloc[0])
        new_line["id"] = new_line["id"].replace(
            new_line["experiment"], "another_experiment"
        )
        new_line["experiment"] = "another_experiment"
        cat.esmcat._df = pd.concat([cat.df, new_line.to_frame().T], ignore_index=True)

        with pytest.warns(
            UserWarning,
            match="doesn't have the fixed field sftlf, but it can be acquired from ",
        ):
            out = xs.search_data_catalogs(
                data_catalogs=cat,
                variables_and_freqs={"sftlf": "fx"},
                other_search_criteria={"experiment": "another_experiment"},
            )
        assert len(out) == 1
        k = list(out.keys())[0]
        np.testing.assert_array_equal(
            out[k].df["experiment"],
            "another_experiment",
        )


class TestGetWarmingLevel:
    def test_list(self):
        with pytest.warns(
            FutureWarning, match="get_warming_level has been deprecated."
        ):
            out = xs.get_warming_level(
                ["CMIP6_CanESM5_ssp126_r1i1p1f1", "CMIP6_CanESM5_ssp245_r1i1p1f1"],
                wl=2,
                window=20,
                return_horizon=False,
            )
        assert isinstance(out, list)
        assert out == ["2026", "2024"]

        period = [int(out[0]) - 9, int(out[0]) + 10]
        out2 = xs.get_warming_level_from_period(
            ["CMIP6_CanESM5_ssp126_r1i1p1f1", "CMIP6_CanESM5_ssp245_r1i1p1f1"],
            period,
        )
        assert isinstance(out2, list)
        np.testing.assert_array_almost_equal(out2, [2.0418553921, 2.1353053921])
        assert (
            xs.get_warming_level_from_period(
                "CMIP6_CanESM5_ssp126_r1i1p1f1", list(np.array(period) - 1)
            )
            < 2
        )

    def test_string_with_horizon(self):
        out = xs.get_period_from_warming_level(
            "CMIP6_CanESM5_ssp585_r1i1p1f1", wl=2, window=20, return_central_year=False
        )
        assert isinstance(out, list)
        assert out == ["2013", "2032"]

        out2 = xs.get_warming_level_from_period("CMIP6_CanESM5_ssp585_r1i1p1f1", out)
        np.testing.assert_array_almost_equal(out2, 2.00547622)

    @pytest.mark.parametrize("attrs", ["global", "regional", "regional_w_institution"])
    def test_ds(self, attrs):
        ds = timeseries(
            np.tile(np.arange(1, 2), 30),
            variable="tas",
            start="2001-01-01",
            freq="YS-JAN",
            as_dataset=True,
        )
        attributes = {
            "global": {
                "cat:mip_era": "CMIP6",
                "cat:source": "CanESM5",
                "cat:experiment": "ssp126",
            },
            "regional": {
                "cat:mip_era": "CMIP6",
                "cat:driving_model": "CanESM5",
                "cat:experiment": "ssp126",
            },
            "regional_w_institution": {
                "cat:mip_era": "CMIP6",
                "cat:driving_institution": "CCCma",
                "cat:driving_model": "CCCma-CanESM5",
                "cat:experiment": "ssp126",
            },
        }

        ds.attrs = attributes[deepcopy(attrs)]
        assert (
            xs.get_period_from_warming_level(
                ds, wl=2, window=20, ignore_member=True, return_central_year=True
            )
            == "2026"
        )
        np.testing.assert_array_almost_equal(
            xs.get_warming_level_from_period(
                ds, [2026 - 9, 2026 + 10], ignore_member=True
            ),
            2.0418553,
        )

    def test_multiple_matches(self):
        # 55 instances of CanESM2-rcp85 in the CSV, but it should still return a single value
        assert (
            xs.get_period_from_warming_level(
                "CMIP5_CanESM2_rcp85_.*", wl=3.5, window=30, return_central_year=True
            )
            == "2059"
        )
        np.testing.assert_array_almost_equal(
            xs.get_warming_level_from_period(
                "CMIP5_CanESM2_rcp85_.*", [2059 - 14, 2059 + 15]
            ),
            3.545244,
        )

    def test_odd_window(self):
        assert xs.get_period_from_warming_level(
            "CMIP6_CanESM5_ssp126_r1i1p1f1", wl=2, window=21, return_central_year=False
        ) == ["2016", "2036"]
        np.testing.assert_array_almost_equal(
            xs.get_warming_level_from_period(
                "CMIP6_CanESM5_ssp126_r1i1p1f1", [2016, 2036]
            ),
            2.019456,
        )

    def test_none(self):
        assert (
            xs.get_period_from_warming_level(
                "CMIP6_CanESM5_ssp585_r1i1p1f1",
                wl=20,
                window=20,
                return_central_year=True,
            )
            is None
        )
        assert (
            xs.get_period_from_warming_level(
                "CMIP6_notreal_ssp585_r1i1p1f1",
                wl=20,
                window=20,
                return_central_year=True,
            )
            is None
        )
        with pytest.raises(ValueError, match="s not fully covered by the provided"):
            xs.get_warming_level_from_period(
                "CMIP6_CanESM5_ssp585_r1i1p1f1", ["2100", "2120"]
            )
        assert (
            xs.get_warming_level_from_period(
                "CMIP6_notreal_ssp585_r1i1p1f1", ["2010", "2020"]
            )
            is None
        )

    def test_wrong_types(self):
        with pytest.raises(ValueError):
            xs.get_period_from_warming_level(
                {"this": "is not valid."}, wl=2, window=20, return_central_year=False
            )
        with pytest.raises(ValueError):
            xs.get_period_from_warming_level(
                "CMIP6_CanESM5_ssp585_r1i1p1f1_toomany_underscores",
                wl=2,
                window=20,
                return_central_year=False,
            )
        with pytest.raises(ValueError):
            xs.get_period_from_warming_level(
                "CMIP6_CanESM5_ssp585_r1i1p1f1",
                wl=2,
                window=3.85,
                return_central_year=False,
            )

    def test_DataArray(self):  # noqa: N802
        reals = xr.DataArray(
            ["CMIP6_CanESM5_ssp126_r1i1p1f1"], dims=("x",), coords={"x": [1]}
        )
        out = xs.get_period_from_warming_level(reals, wl=2, return_central_year=True)
        xr.testing.assert_identical(out, reals.copy(data=["2026"]))

        out2 = xs.get_warming_level_from_period(reals, [2026 - 9, 2026 + 10])
        np.testing.assert_array_almost_equal(out2, [2.0418553921])

    def test_DataFrame(self):  # noqa: N802
        reals = pd.DataFrame.from_records(
            [
                {
                    "mip_era": "CMIP6",
                    "source": "CanESM5",
                    "experiment": "ssp126",
                    "member": "r1i1p1f1",
                },
                {
                    "mip_era": "CMIP6",
                    "source": "CanESM5",
                    "experiment": "ssp245",
                    "member": "r1i1p1f1",
                },
            ],
            index=["a", "b"],
        )
        out = xs.get_period_from_warming_level(
            reals,
            wl=2,
            window=20,
            return_central_year=True,
        )
        pd.testing.assert_series_equal(
            out, pd.Series(["2026", "2024"], index=["a", "b"])
        )

        out2 = xs.get_warming_level_from_period(reals, [2026 - 9, 2026 + 10])
        np.testing.assert_array_almost_equal(out2, [2.0418553921, 2.1353053921])


class TestSubsetWarmingLevel:
    ds = timeseries(
        np.tile(np.arange(1, 2), 50),
        variable="tas",
        start="2000-01-01",
        freq="YS-JAN",
        as_dataset=True,
    ).assign_attrs(
        {
            "cat:mip_era": "CMIP6",
            "cat:source": "CanESM5",
            "cat:experiment": "ssp585",
            "cat:member": "r1i1p1f1",
        }
    )

    def test_default(self):
        ds_sub = xs.subset_warming_level(self.ds, wl=2)

        np.testing.assert_array_equal(ds_sub.time.dt.year[0], 2013)
        np.testing.assert_array_equal(ds_sub.time.dt.year[-1], 2032)
        np.testing.assert_array_equal(ds_sub.warminglevel[0], "+2Cvs1850-1900")
        assert ds_sub.warminglevel.attrs["baseline"] == "1850-1900"
        assert ds_sub.attrs["cat:processing_level"] == "warminglevel-2vs1850-1900"

    def test_kwargs(self):
        ds_sub = xs.subset_warming_level(
            self.ds,
            wl=1,
            window=25,
            tas_baseline_period=["1981", "2010"],
            to_level="tests",
        )

        np.testing.assert_array_equal(ds_sub.time.dt.year[0], 2009)
        np.testing.assert_array_equal(ds_sub.time.dt.year[-1], 2033)
        np.testing.assert_array_equal(ds_sub.warminglevel[0], "+1Cvs1981-2010")
        assert ds_sub.warminglevel.attrs["baseline"] == "1981-2010"
        assert ds_sub.attrs["cat:processing_level"] == "tests"

    @pytest.mark.parametrize("wl", [3.5, 5, 20, [2, 3.5, 5, 20], [3.5, 5, 20]])
    def test_outofrange(self, wl):
        # 3.5 is only partially covered by ds, 5 is out of range but within the csv, 20 is fully out of range
        if not isinstance(wl, list):
            assert xs.subset_warming_level(self.ds, wl=wl) is None
        else:
            ds = xs.subset_warming_level(self.ds, wl=wl)
            assert ds.warminglevel.size == len(wl)
            if len(wl) == 3:
                np.testing.assert_array_equal(ds.tas.isnull().all(), [True])
            else:
                np.testing.assert_array_equal(
                    ds.tas.isnull().all(dim="time"), [False, True, True, True]
                )

    def test_multireals(self):
        ds = self.ds.expand_dims(
            realization=[
                "CMIP6_CanESM5_ssp126_r1i1p1f1",
                "CMIP6_CanESM5_ssp245_r1i1p1f1",
                "fake_faux_falsch_falso",
            ]
        )
        ds_sub = xs.subset_warming_level(
            ds,
            wl=1.5,
            to_level="tests",
        )
        np.testing.assert_array_equal(ds_sub.time.dt.year, np.arange(1000, 1020))
        np.testing.assert_array_equal(
            ds_sub.warminglevel_bounds[:2].dt.year, [[[2004, 2023]], [[2004, 2023]]]
        )
        assert ds_sub.warminglevel_bounds[2].isnull().all()

    def test_multilevels(self):
        ds_sub = xs.subset_warming_level(
            self.ds,
            wl=[1, 2, 3, 20],
            to_level="tests",
        )
        np.testing.assert_array_equal(
            ds_sub.warminglevel,
            ["+1Cvs1850-1900", "+2Cvs1850-1900", "+3Cvs1850-1900", "+20Cvs1850-1900"],
        )
        np.testing.assert_array_equal(ds_sub.tas.isnull().sum("time"), [20, 0, 20, 20])


class TestResample:
    @pytest.mark.parametrize(
        "infrq,meth,outfrq,exp",
        [
            ["MS", "mean", "QS-DEC", [0.47457627, 3, 6.01086957, 9, 11.96666667]],
            [
                "QS-DEC",
                "mean",
                "YS",
                [1.49041096, 5.49041096, 9.49453552, 13.49041096, 17.49041096],
            ],
            ["MS", "std", "2YS", [6.92009239, 6.91557206]],
            [
                "QS",
                "median",
                "YS",
                [1.516437, 5.516437, 9.516437, 13.51092864, 17.516437],
            ],
        ],
    )
    def test_weighted(self, infrq, meth, outfrq, exp):
        da = timeseries(
            np.arange(48),
            variable="tas",
            start="2001-01-01",
            freq=infrq,
        )
        out = xs.extract.resample(da, outfrq, method=meth)
        np.testing.assert_allclose(out.isel(time=slice(0, 5)), exp)

    def test_weighted_wind(self):
        uas = timeseries(
            np.arange(48),
            variable="uas",
            start="2001-01-01",
            freq="MS",
        )
        vas = timeseries(
            np.arange(48),
            variable="vas",
            start="2001-01-01",
            freq="MS",
        )
        ds = xr.merge([uas, vas])
        out = xs.extract.resample(ds.uas, "YS", method="wind_direction", ds=ds)
        np.testing.assert_allclose(out, [5.5260274, 17.5260274, 29.5260274, 41.5136612])

    def test_missing(self):
        da = timeseries(
            np.arange(48),
            variable="tas",
            start="2001-01-01",
            freq="MS",
        )
        out = xs.extract.resample(da, "QS-DEC", method="mean", missing="drop")
        assert out.size == 15

        out = xs.extract.resample(da, "QS-DEC", method="mean", missing="mask")
        assert out.isel(time=0).isnull().all()

    def test_missing_xclim(self):
        arr = np.arange(48).astype(float)
        arr[0] = np.nan
        arr[40:] = np.nan
        da = timeseries(
            arr,
            variable="tas",
            start="2001-01-01",
            freq="MS",
        )
        out = xs.extract.resample(da, "YS", method="mean", missing={"method": "any"})
        assert out.isel(time=0).isnull().all()

        out = xs.extract.resample(
            da, "YS", method="mean", missing={"method": "pct", "tolerance": 0.6}
        )
        assert out.isel(time=0).notnull().all()
        assert out.isel(time=-1).isnull().all()



import geopandas as gpd
import numpy as np
import pytest
import xarray as xr
import xclim
from conftest import notebooks
from shapely.geometry import Polygon
from xclim.testing.helpers import test_timeseries as timeseries

import xscen as xs
from xscen.testing import datablock_3d

try:
    import xesmf as xe
except ImportError:
    xe = None


class TestComputeDeltas:
    ds = xs.climatological_op(
        timeseries(
            np.repeat(np.arange(1, 5), 30).astype(float),
            variable="tas",
            start="1981-01-01",
            freq="YS",
            as_dataset=True,
        ),
        op="mean",
        window=30,
        stride=30,
        rename_variables=False,
    )

    @pytest.mark.parametrize(
        "kind, rename_variables, to_level",
        [("+", True, None), ("/", False, "for_testing"), ("%", True, "for_testing")],
    )
    def test_options(self, kind, rename_variables, to_level):
        if to_level is None:
            deltas = xs.compute_deltas(
                self.ds,
                reference_horizon="1981-2010",
                kind=kind,
                rename_variables=rename_variables,
            )
        else:
            deltas = xs.compute_deltas(
                self.ds,
                reference_horizon="1981-2010",
                kind=kind,
                rename_variables=rename_variables,
                to_level=to_level,
            )

        variable = "tas_delta_1981_2010" if rename_variables else "tas"
        delta_kind = "abs." if kind == "+" else "rel." if kind == "/" else "pct."
        results = (
            [0, 1, 2, 3]
            if kind == "+"
            else [1, 2, 3, 4] if kind == "/" else [0, 100, 200, 300]
        )
        units = "K" if kind == "+" else "" if kind == "/" else "%"

        # Test rename_variables and to_level
        assert variable in deltas.data_vars
        assert len(deltas.data_vars) == 1
        assert deltas.attrs["cat:processing_level"] == (
            "for_testing" if to_level is not None else "deltas"
        )
        # Test metadata
        assert (
            deltas[variable].attrs["description"]
            == f"{self.ds.tas.attrs['description'].strip(' .')}: {delta_kind} delta compared to 1981-2010."
        )
        assert f"{delta_kind} delta vs. 1981-2010" in deltas[variable].attrs["history"]
        # Test variable
        assert deltas[variable].attrs["delta_kind"] == delta_kind
        assert deltas[variable].attrs["delta_reference"] == "1981-2010"
        assert deltas[variable].attrs["units"] == units
        if kind == "+":
            assert deltas[variable].attrs["units_metadata"] == "temperature: difference"
        np.testing.assert_array_equal(deltas[variable], results)

    def test_ref_ds(self):
        ref = self.ds.where(self.ds.horizon == "1981-2010", drop=True)
        out = xs.compute_deltas(self.ds, reference_horizon=ref)
        np.testing.assert_array_equal(out["tas_delta_1981_2010"], np.arange(0, 4))

        ref = ref.drop_vars("horizon")
        out = xs.compute_deltas(self.ds, reference_horizon=ref)
        np.testing.assert_array_equal(out["tas_delta_unknown_horizon"], np.arange(0, 4))

    @pytest.mark.parametrize("cal", ["proleptic_gregorian", "noleap", "360_day"])
    def test_calendars(self, cal):
        out = xs.compute_deltas(
            self.ds.convert_calendar(cal, align_on="date"),
            reference_horizon="1981-2010",
        )
        assert out.time.dt.calendar == cal

    def test_input_ds(self):
        out1 = xs.compute_deltas(
            self.ds,
            reference_horizon=self.ds.where(self.ds.horizon == "1981-2010", drop=True),
        )
        out2 = xs.compute_deltas(self.ds, reference_horizon="1981-2010")
        assert out1.equals(out2)

    @pytest.mark.parametrize("xrfreq", ["MS", "QS", "YS-JAN"])
    def test_freqs(self, xrfreq):
        o = 12 if xrfreq == "MS" else 4 if xrfreq == "QS" else 1

        ds = xs.climatological_op(
            timeseries(
                np.repeat(np.arange(1, 5), 30 * o).astype(float),
                variable="tas",
                start="1981-01-01",
                freq=xrfreq,
                as_dataset=True,
            ),
            op="mean",
            window=30,
            stride=30,
            rename_variables=False,
        )

        out = xs.compute_deltas(
            ds, reference_horizon="1981-2010", rename_variables=False
        )
        assert len(out.time) == o * len(np.unique(out.horizon.values))
        results = np.repeat(np.arange(0, 4), o)
        np.testing.assert_array_equal(out.tas, results)

    def test_fx(self):
        ds = timeseries(
            np.arange(1, 5),
            variable="tas",
            start="1981-01-01",
            freq="30YS",
            as_dataset=True,
        )
        ds["horizon"] = xr.DataArray(
            ["1981-2010", "2011-2040", "2041-2070", "2071-2100"], dims=["time"]
        )
        ds = ds.swap_dims({"time": "horizon"}).drop_vars("time")

        out = xs.compute_deltas(
            ds, reference_horizon="1981-2010", rename_variables=False
        )
        np.testing.assert_array_equal(out.tas, np.arange(0, 4))
        out2 = xs.compute_deltas(
            ds,
            reference_horizon=ds.sel(horizon="1981-2010"),
            kind="+",
            rename_variables=False,
        )
        assert out.equals(out2)

    def test_errors(self):
        # Multiple horizons in reference
        with pytest.raises(ValueError):
            xs.compute_deltas(
                self.ds,
                reference_horizon=self.ds.where(
                    self.ds.horizon.isin(["1981-2010", "2011-2040"]), drop=True
                ),
            )
        # Unknown reference horizon
        with pytest.raises(ValueError):
            xs.compute_deltas(self.ds, reference_horizon="1981-2010-2030")
        with pytest.raises(ValueError):
            xs.compute_deltas(self.ds, reference_horizon=5)
        # Unknown reference horizon format
        with pytest.raises(ValueError):
            xs.compute_deltas(
                self.ds,
                reference_horizon=self.ds.where(
                    self.ds.horizon == "1981-2010", drop=True
                )["tas"],
            )
        # Unknown kind
        with pytest.raises(ValueError):
            xs.compute_deltas(self.ds, reference_horizon="1981-2010", kind="unknown")
        # Daily data
        ds = timeseries(
            np.tile(np.arange(0, 365), 3),
            variable="tas",
            start="2001-01-01",
            freq="D",
            as_dataset=True,
        )
        ds["horizon"] = ds.time.dt.year.astype(str)
        with pytest.raises(NotImplementedError):
            xs.compute_deltas(ds, reference_horizon="2001")


class TestProduceHorizon:
    ds = timeseries(
        np.ones(365 * 30 + 7),
        variable="tas",
        start="1981-01-01",
        freq="D",
        as_dataset=True,
    )
    ds.attrs["cat:source"] = "CanESM5"
    ds.attrs["cat:experiment"] = "ssp585"
    ds.attrs["cat:member"] = "r1i1p1f1"
    ds.attrs["cat:mip_era"] = "CMIP6"

    yaml_file = notebooks / "samples" / "indicators.yml"

    @pytest.mark.parametrize(
        "periods, to_level",
        [
            (None, None),
            ([["1995", "2007"], ["1995", "1996"]], "for_testing"),
        ],
    )
    def test_options(self, periods, to_level):
        if to_level is None:
            out = xs.produce_horizon(
                self.ds, indicators=self.yaml_file, periods=periods
            )
        else:
            with pytest.warns(UserWarning, match="The attributes for variable tg_min"):
                out = xs.produce_horizon(
                    self.ds,
                    indicators=self.yaml_file,
                    periods=periods,
                    to_level=to_level,
                )

        assert "time" not in out.dims
        assert len(out.horizon) == 1 if periods is None else len(periods)
        np.testing.assert_array_equal(
            out.horizon,
            ["1981-2010"] if periods is None else ["1995-2007", "1995-1996"],
        )
        assert out.attrs["cat:processing_level"] == (
            "for_testing" if to_level is not None else "horizons"
        )
        assert out.attrs["cat:xrfreq"] == "fx"
        assert all(v in out for v in ["tg_min", "growing_degree_days"])
        assert (
            f"{30 if periods is None else int(periods[0][1]) - int(periods[0][0]) + 1}-year climatological average of"
            in out.tg_min.attrs["description"]
        )
        assert (
            out.tg_min.attrs["description"].split(
                f"{30 if periods is None else int(periods[0][1]) - int(periods[0][0]) + 1}"
                f"-year climatological average of "
            )[1]
            != self.ds.tas.attrs["description"]
        )
        np.testing.assert_array_equal(out.tg_min, [1] * len(out.horizon))
        np.testing.assert_array_equal(out.growing_degree_days, [0] * len(out.horizon))

    def test_multiple_freqs(self):
        ds = timeseries(
            np.ones(365 * 30 + 7),
            variable="tas",
            start="1981-01-01",
            freq="D",
            as_dataset=True,
        )
        ds["tas"] = ds["tas"].convert_calendar("noleap", align_on="date")
        # Overwrite values to make them equal to the month
        ds["tas"].values = ds["time"].dt.month
        ds["da"] = ds["tas"]

        indicator_qs = xclim.core.indicator.Indicator.from_dict(
            data={"base": "tg_min", "parameters": {"freq": "QS-DEC"}},
            identifier="tg_min_qs",
            module="tests",
        )
        indicator_ms = xclim.core.indicator.Indicator.from_dict(
            data={"base": "tg_min", "parameters": {"freq": "MS"}},
            identifier="tg_min_ms",
            module="tests",
        )

        indicators = [
            ("fit", xclim.indicators.generic.fit),
            ("tg_min_as", xclim.indicators.atmos.tg_min),
            ("tg_min_qs", indicator_qs),
            ("tg_min_ms", indicator_ms),
        ]

        out = xs.produce_horizon(ds, indicators=indicators)
        assert len(out.horizon) == 1
        assert all(v in out for v in ["params", "tg_min", "tg_min_qs", "tg_min_ms"])
        np.testing.assert_array_equal(out["season"], ["MAM", "JJA", "SON", "DJF"])
        np.testing.assert_array_equal(
            out["month"],
            [
                "JAN",
                "FEB",
                "MAR",
                "APR",
                "MAY",
                "JUN",
                "JUL",
                "AUG",
                "SEP",
                "OCT",
                "NOV",
                "DEC",
            ],
        )
        np.testing.assert_array_almost_equal(
            out["params"].squeeze(), [6.523, 3.449], decimal=3
        )
        np.testing.assert_array_equal(out["tg_min"].squeeze(), [1])
        np.testing.assert_array_equal(out["tg_min_qs"].squeeze(), [3, 6, 9, 1])
        np.testing.assert_array_equal(out["tg_min_ms"].squeeze(), np.arange(1, 13))

    @pytest.mark.parametrize("wl", [0.8, [0.8, 0.85]])
    def test_warminglevels(self, wl):
        out = xs.produce_horizon(
            self.ds, indicators=self.yaml_file, warminglevels={"wl": wl}
        )
        assert "warminglevel" not in out.dims
        assert len(out.horizon) == 1 if isinstance(wl, float) else len(wl)
        np.testing.assert_array_equal(
            out.horizon,
            (
                ["+0.8Cvs1850-1900"]
                if isinstance(wl, float)
                else ["+0.8Cvs1850-1900", "+0.85Cvs1850-1900"]
            ),
        )

    def test_combine(self):
        out = xs.produce_horizon(
            self.ds,
            indicators=self.yaml_file,
            periods=[["1982", "1988"]],
            warminglevels={"wl": [0.8, 0.85]},
        )
        assert len(out.horizon) == 3
        np.testing.assert_array_equal(
            out.horizon, ["1982-1988", "+0.8Cvs1850-1900", "+0.85Cvs1850-1900"]
        )

    def test_single(self):
        out = xs.produce_horizon(
            self.ds,
            indicators=self.yaml_file,
            periods=[1982, 1988],
        )
        assert len(out.horizon) == 1
        np.testing.assert_array_equal(out.horizon, ["1982-1988"])

    def test_op(self):
        ds = self.ds.copy()
        ds.tas.loc["1995-01-01":"1995-12-31"] = 2
        out = xs.produce_horizon(
            ds, indicators=self.yaml_file, periods=[["1995", "2007"]]
        )
        np.testing.assert_array_almost_equal(out["tg_min"], [14 / 13])
        out2 = xs.produce_horizon(
            ds, indicators=self.yaml_file, periods=[["1995", "2007"]], op="max"
        )
        np.testing.assert_array_almost_equal(out2["tg_min"], [2])

    def test_precomputed(self):
        ds = timeseries(
            np.arange(30),
            variable="tas",
            start="1981-01-01",
            freq="YS-JAN",
            as_dataset=True,
        )
        ds = ds.rename({"tas": "tg_min"})
        ds.attrs["cat:source"] = "CanESM5"
        ds.attrs["cat:experiment"] = "ssp585"
        ds.attrs["cat:member"] = "r1i1p1f1"
        ds.attrs["cat:mip_era"] = "CMIP6"

        out = xs.produce_horizon(ds, periods=[["1981", "1985"], ["1991", "1995"]])
        assert len(out.horizon) == 2
        assert "time" not in out.dims
        np.testing.assert_array_almost_equal(
            out["tg_min"], [np.mean(np.arange(5)), np.mean(np.arange(10, 15))]
        )

    def test_warminglevel_in_ds(self):
        ds = self.ds.copy().expand_dims({"warminglevel": ["+1Cvs1850-1900"]})
        out = xs.produce_horizon(
            ds, indicators=self.yaml_file, to_level="warminglevel{wl}"
        )
        np.testing.assert_array_equal(out["horizon"], ds["warminglevel"])
        assert out.attrs["cat:processing_level"] == "warminglevel+1Cvs1850-1900"

        # Multiple warming levels
        ds = self.ds.copy().expand_dims(
            {"warminglevel": ["+1Cvs1850-1900", "+2Cvs1850-1900"]}
        )
        with pytest.raises(ValueError):
            xs.produce_horizon(ds, indicators=self.yaml_file)

    def test_to_level(self):
        out = xs.produce_horizon(
            self.ds, indicators=self.yaml_file, to_level="horizon{period0}-{period1}"
        )
        assert out.attrs["cat:processing_level"] == "horizon1981-2010"
        out = xs.produce_horizon(
            self.ds,
            indicators=self.yaml_file,
            warminglevels={"wl": 1, "tas_baseline_period": ["1851", "1901"]},
            to_level="warminglevel{wl}",
        )
        assert out.attrs["cat:processing_level"] == "warminglevel+1Cvs1851-1901"

    def test_errors(self):
        # Bad input
        with pytest.raises(
            ValueError, match="Could not understand the format of warminglevels"
        ):
            xs.produce_horizon(
                self.ds, indicators=self.yaml_file, warminglevels={"wl": "+1"}
            )

        # Insufficient data
        with pytest.warns(
            UserWarning, match="is not fully covered by the input dataset."
        ):
            with pytest.raises(
                ValueError, match="No horizon could be computed. Check your inputs."
            ):
                xs.produce_horizon(
                    self.ds, indicators=self.yaml_file, periods=[["1982", "2100"]]
                )
        with pytest.warns(
            UserWarning, match="is not fully covered by the input dataset."
        ):
            with pytest.raises(
                ValueError, match="No horizon could be computed. Check your inputs."
            ):
                xs.produce_horizon(
                    self.ds, indicators=self.yaml_file, periods=[["1950", "1990"]]
                )


class TestSpatialMean:
    # We test different longitude flavors : all < 0, crossing 0, all > 0
    # the default global bbox changes because of subtleties in clisops
    @pytest.mark.parametrize(
        "method,exp", (["xesmf", 1.62032976], ["cos-lat", 1.63397460])
    )
    @pytest.mark.parametrize("lonstart", [-70, -30, 0])
    def test_global(self, lonstart, method, exp):
        if method == "xesmf" and xe is None:
            pytest.skip("xesmf needed for testing averaging with method xesmf")
        ds = datablock_3d(
            np.array([[[0, 1, 2], [1, 2, 3], [2, 3, 4]]] * 3, "float"),
            "tas",
            "lon",
            lonstart,
            "lat",
            15,
            30,
            30,
            as_dataset=True,
        )

        avg = xs.aggregate.spatial_mean(ds, method=method, region="global")
        np.testing.assert_allclose(avg.tas, exp)

    def test_coslat_2d(self):
        ds = datablock_3d(
            np.array([[[0, 1, 2], [1, 2, 3], [2, 3, 4]]] * 3, "float"),
            "tas",
            "rlon",
            0,
            "rlat",
            15,
            30,
            30,
            as_dataset=True,
        )
        out = xs.aggregate.spatial_mean(ds, method="cos-lat")
        assert "rotated_pole" not in out
        assert "grid_mapping" not in out.tas.attrs
        np.testing.assert_allclose(out.tas, 1.623069, rtol=1e-6)
        assert "weighted mean(dim=('rlat', 'rlon'))" in out.attrs["history"]

    def test_cos_lat_errors(self):
        ds = datablock_3d(
            np.array([[[0, 1, 2], [1, 2, 3], [2, 3, 4]]] * 3, "float"),
            "tas",
            "lon",
            -70,
            "lat",
            15,
            30,
            30,
            as_dataset=True,
        )

        ds_no_attrs = ds.copy()
        ds_no_attrs["lat"].attrs = {}
        ds_no_attrs["lon"].attrs = {}
        out0 = xs.aggregate.spatial_mean(
            ds_no_attrs, method="cos-lat"
        )  # Should not raise, since 'lat' is in the coords
        with pytest.raises(
            ValueError, match="Could not determine the spatial coordinate names"
        ):
            xs.aggregate.spatial_mean(
                ds_no_attrs.rename({"lat": "some_other_name"}), method="cos-lat"
            )
        out1 = xs.aggregate.spatial_mean(
            ds_no_attrs.rename({"lat": "some_other_name"}),
            method="cos-lat",
            kwargs={"latitude_name": "some_other_name"},
        )
        np.testing.assert_array_almost_equal(out0.tas, out1.tas)

        ds_no_units = ds.copy()
        ds_no_units["lat"].attrs.pop("units")
        with pytest.warns(UserWarning, match="Latitude units are "):
            out2 = xs.aggregate.spatial_mean(ds_no_units, method="cos-lat")
        np.testing.assert_array_almost_equal(out0.tas, out2.tas)

        ds2 = datablock_3d(
            np.array([[[0, 1, 2], [1, 2, 3], [2, 3, 4]]] * 3, "float"),
            "tas",
            "lon",
            170,
            "lat",
            15,
            10,
            30,
            as_dataset=True,
        )
        with xr.set_options(keep_attrs=True):
            ds2["lon"] = xr.where(ds2["lon"] > 180, ds2["lon"] - 360, ds2["lon"])
        with pytest.warns(
            UserWarning, match="The region appears to be crossing the -180/180"
        ):
            xs.aggregate.spatial_mean(ds2, method="cos-lat")

    def test_interp(self):
        ds = datablock_3d(
            np.array([[[0, 1, 2], [1, 2, 3], [2, 3, 4]]] * 3, "float"),
            "tas",
            "rlon",
            0,
            "rlat",
            15,
            30,
            30,
            as_dataset=True,
        )

        with pytest.raises(
            ValueError, match="has been removed, since it produced incorrect results."
        ):
            xs.aggregate.spatial_mean(ds, method="interp_centroid")

    def test_xesmf_shape(self):
        if xe is None:
            pytest.skip("xesmf needed for testing averaging with method xesmf")
        ds = datablock_3d(
            np.array([[[0, 1, 2], [1, 2, 3], [2, 3, 4]]] * 3, "float"),
            "tas",
            "lon",
            -70,
            "lat",
            45,
            30,
            10,
            as_dataset=True,
        )
        poly = gpd.GeoDataFrame(
            geometry=[Polygon([(-60, 50), (-35, 50), (-35, 60), (-60, 60)])]
        )
        region = {"method": "shape", "shape": poly}

        avg = xs.aggregate.spatial_mean(ds, method="xesmf", region=region)
        np.testing.assert_allclose(avg.tas, 1.729156)
        assert avg.attrs["regrid_method"] == "conservative"
        assert "xesmf.SpatialAverager over 1 polygon" in avg.attrs["history"]

    @pytest.mark.parametrize("simplify_tolerance", [None, 50])
    def test_xesmf_rot_shape(self, simplify_tolerance):
        if xe is None:
            pytest.skip("xesmf needed for testing averaging with method xesmf")
        ds = datablock_3d(
            np.array([[[0, 1, 2], [1, 2, 3], [2, 3, 4]]] * 3, "float"),
            "tas",
            "rlon",
            0,
            "rlat",
            45,
            30,
            10,
            as_dataset=True,
        )
        poly = gpd.GeoDataFrame(
            geometry=[Polygon([(60, 50), (30, 50), (30, 60), (60, 60)])]
        )
        region = {"method": "shape", "shape": poly}

        avg = xs.aggregate.spatial_mean(
            ds,
            method="xesmf",
            region=region,
            simplify_tolerance=simplify_tolerance,
            to_level="for_testing",
            to_domain="for_testing2",
        )
        if simplify_tolerance is None:
            np.testing.assert_allclose(avg.tas, 3.302546, rtol=1e-6)
            assert avg.attrs["regrid_method"] == "conservative"
            assert "rotated_pole" not in avg
            assert "grid_mapping" not in avg.tas.attrs
            assert all(c in avg.coords for c in ["lon_bounds", "lat_bounds"])
            np.testing.assert_array_equal(avg["lon"], np.mean([60, 30]))
            np.testing.assert_array_equal(avg["lat"], np.mean([50, 60]))
            assert avg.attrs["cat:processing_level"] == "for_testing"
            assert avg.attrs["cat:domain"] == "for_testing2"
        else:
            # Essentially just test that it changes the results
            np.testing.assert_allclose(avg.tas, 2.967405, rtol=1e-6)
            np.testing.assert_array_almost_equal(avg["lon"], 40)
            np.testing.assert_array_almost_equal(avg["lat"], 56.66666667)

    def test_errors(self):
        if xe is None:
            pytest.skip("xesmf needed for testing averaging with method xesmf")
        ds = datablock_3d(
            np.array([[[0, 1, 2], [1, 2, 3], [2, 3, 4]]] * 3, "float"),
            "tas",
            "lon",
            -70,
            "lat",
            45,
            30,
            10,
            as_dataset=True,
        )
        with pytest.raises(ValueError, match="should be one of"):
            xs.aggregate.spatial_mean(ds, method="xesmf", region={"method": "foo"})
        with pytest.raises(ValueError, match="Subsetting method should be"):
            xs.aggregate.spatial_mean(ds, method="foo")


class TestClimatologicalOp:
    @staticmethod
    def _format(s):
        import xclim

        op_format = dict.fromkeys(("mean", "std", "var", "sum"), "adj") | dict.fromkeys(
            ("max", "min"), "noun"
        )
        return xclim.core.formatting.default_formatter.format_field(s, op_format[s])

    def test_daily(self):
        ds = timeseries(
            np.tile(np.arange(1, 13), 3),
            variable="tas",
            start="2001-01-01",
            freq="D",
            as_dataset=True,
        )
        with pytest.raises(NotImplementedError):
            xs.climatological_op(ds, op="mean")

    @pytest.mark.parametrize("xrfreq", ["MS", "YS-JAN"])
    @pytest.mark.parametrize(
        "op", ["max", "mean", "median", "min", "std", "sum", "var", "linregress"]
    )
    def test_all_default(self, xrfreq, op):
        o = 12 if xrfreq == "MS" else 1

        ds = timeseries(
            np.tile(np.arange(1, o + 1), 30),
            variable="tas",
            start="2001-01-01",
            freq=xrfreq,
            as_dataset=True,
        )
        out = xs.climatological_op(ds, op=op)
        expected = (
            dict.fromkeys(("max", "mean", "median", "min"), np.arange(1, o + 1))
            | dict.fromkeys(("std", "var"), np.zeros(o))
            | dict({"sum": np.arange(1, o + 1) * 30})
            | dict(
                {
                    "linregress": np.array(
                        [
                            np.zeros(o),
                            np.arange(1, o + 1),
                            np.zeros(o),
                            np.ones(o),
                            np.zeros(o),
                            np.zeros(o),
                        ]
                    ).T
                }
            )
        )
        # Test output variable name, values, length, horizon
        assert list(out.data_vars.keys()) == [f"tas_clim_{op}"]
        np.testing.assert_array_equal(out[f"tas_clim_{op}"], expected[op])
        assert len(out.time) == (o * len(np.unique(out.horizon.values)))
        np.testing.assert_array_equal(out.time[0], ds.time[0])
        assert (out.horizon == "2001-2030").all()
        # Test metadata
        operation = self._format(op) if op not in ["median", "linregress"] else op
        assert (
            out[f"tas_clim_{op}"].attrs["description"]
            == f"30-year climatological {operation} of {ds.tas.attrs['description']}"
        )
        assert (
            f"30-year climatological {operation} over window (non-centered), with a minimum of 30 years of data"
            in out[f"tas_clim_{op}"].attrs["history"]
        )
        assert out.attrs["cat:processing_level"] == "climatology"

    @pytest.mark.parametrize("xrfreq", ["MS", "YS-JAN"])
    @pytest.mark.parametrize(
        "op", ["max", "mean", "median", "min", "std", "sum", "var", "linregress"]
    )
    def test_options(self, xrfreq, op):
        o = 12 if xrfreq == "MS" else 1

        ds = timeseries(
            np.tile(np.arange(1, o + 1), 30),
            variable="tas",
            start="2001-01-01",
            freq=xrfreq,
            as_dataset=True,
        )
        out = xs.climatological_op(
            ds, op=op, window=15, stride=5, to_level="for_testing"
        )
        expected = (
            dict.fromkeys(
                ("max", "mean", "median", "min"),
                np.tile(np.arange(1, o + 1), len(np.unique(out.horizon.values))),
            )
            | dict.fromkeys(
                ("std", "var"), np.tile(np.zeros(o), len(np.unique(out.horizon.values)))
            )
            | dict(
                {
                    "sum": np.tile(
                        np.arange(1, o + 1) * 15, len(np.unique(out.horizon.values))
                    )
                }
            )
            | dict(
                {
                    "linregress": np.tile(
                        np.array(
                            [
                                np.zeros(o),
                                np.arange(1, o + 1),
                                np.zeros(o),
                                np.ones(o),
                                np.zeros(o),
                                np.zeros(o),
                            ]
                        ),
                        len(np.unique(out.horizon.values)),
                    ).T
                }
            )
        )
        # Test output values
        np.testing.assert_array_equal(
            out[f"tas_clim_{op}"],
            expected[op],
        )
        assert len(out.time) == (o * len(np.unique(out.horizon.values)))
        np.testing.assert_array_equal(out.time[0], ds.time[0])
        assert {"2001-2015", "2006-2020", "2011-2025", "2016-2030"}.issubset(
            out.horizon.values
        )
        # Test metadata
        operation = self._format(op) if op not in ["median", "linregress"] else op
        assert (
            out[f"tas_clim_{op}"].attrs["description"]
            == f"15-year climatological {operation} of {ds.tas.attrs['description']}"
        )
        assert (
            f"15-year climatological {operation} over window (non-centered), with a minimum of 15 years of data"
            in out[f"tas_clim_{op}"].attrs["history"]
        )
        assert out.attrs["cat:processing_level"] == "for_testing"

    @pytest.mark.parametrize("ddof", [0, 1])
    def test_dict(self, ddof):
        ds = timeseries(
            np.arange(1, 31),
            variable="tas",
            start="2001-01-01",
            freq="YS",
            as_dataset=True,
        )
        ds["tas"] = ds["tas"].astype(float)
        ds["tas"].values[0] = np.nan
        op = {"std": {"ddof": ddof}}

        # Test with NaNs
        with pytest.warns(RuntimeWarning, match="Degrees of freedom <= 0 for slice."):
            np.testing.assert_array_almost_equal(
                xs.climatological_op(ds, op=op).tas_clim_std, np.nan
            )

            out = xs.climatological_op(ds, op=op, min_periods=29)

        if ddof == 0:
            np.testing.assert_array_almost_equal(
                out["tas_clim_std"], 8.366600, decimal=6
            )
            np.testing.assert_array_almost_equal(
                out["tas_clim_std"],
                xs.climatological_op(ds, op="std", min_periods=29)["tas_clim_std"],
            )
        else:
            np.testing.assert_array_almost_equal(
                out["tas_clim_std"], 8.514693, decimal=6
            )

    def test_op_not_implemented(self):
        ds = timeseries(
            np.arange(1, 31),
            variable="tas",
            start="2001-01-01",
            freq="YS",
            as_dataset=True,
        )
        with pytest.warns(UserWarning, match="has not been tested"):
            xs.climatological_op(ds, op="argmax")

    def test_std_mean(self):
        ds = timeseries(
            np.arange(1, 31),
            variable="tas",
            start="2001-01-01",
            freq="YS",
            as_dataset=True,
        )
        ds = ds.rename({"tas": "tas_std"})

        out = xs.climatological_op(ds, op="mean")
        np.testing.assert_array_almost_equal(
            out["tas_std_clim_mean"], np.sqrt(np.square(np.arange(1, 31)).mean())
        )

    @pytest.mark.parametrize("op", ["mean", "linregress"])
    def test_minperiods(self, op):
        ds = timeseries(
            np.tile(np.arange(1, 5), 30),
            variable="tas",
            start="2001-03-01",
            freq="QS-DEC",
            as_dataset=True,
        )
        ds = ds.where(ds["time"].dt.strftime("%Y-%m-%d") != "2030-12-01")

        out = xs.climatological_op(ds, op=op, window=30)
        assert np.sum(np.isnan(out[f"tas_clim_{op}"])) == 0
        assert len(out.time) == 4
        if op == "mean":
            np.testing.assert_array_equal(out[f"tas_clim_{op}"], np.arange(1, 5))
        else:
            np.testing.assert_array_almost_equal(
                out[f"tas_clim_{op}"],
                np.array(
                    [
                        [0.0, 1.0, 0.0, 1.0, 0.0, 0.0],
                        [0.0, 2.0, 0.0, 1.0, 0.0, 0.0],
                        [0.0, 3.0, 0.0, 1.0, 0.0, 0.0],
                        [0.0, 4.0, 0.0, 1.0, 0.0, 0.0],
                    ]
                ),
            )

        # min_periods as int
        out = xs.climatological_op(ds, op=op, window=30, min_periods=30)
        assert np.sum(np.isnan(out[f"tas_clim_{op}"])) == 1 if op == "mean" else 6

        # min_periods as float
        out = xs.climatological_op(ds, op=op, window=30, min_periods=0.5)
        assert "minimum of 15 years of data" in out[f"tas_clim_{op}"].attrs["history"]
        assert np.sum(np.isnan(out[f"tas_clim_{op}"])) == 0

        with pytest.raises(ValueError):
            xs.climatological_op(ds, op=op, window=5, min_periods=6)

        with pytest.raises(ValueError, match="it must be between 0 and 1"):
            xs.climatological_op(ds, op=op, window=30, min_periods=6.5)

    @pytest.mark.parametrize("op", ["mean", "linregress"])
    def test_periods(self, op):
        ds1 = timeseries(
            np.tile(np.arange(1, 2), 10),
            variable="tas",
            start="2001-01-01",
            freq="YS-JAN",
            as_dataset=True,
        )
        ds2 = timeseries(
            np.tile(np.arange(1, 2), 10),
            variable="tas",
            start="2021-01-01",
            freq="YS-JAN",
            as_dataset=True,
        )

        ds = xr.concat([ds1, ds2], dim="time")
        with pytest.raises(ValueError):
            xs.climatological_op(ds, op=op)

        out = xs.climatological_op(
            ds, op="mean", periods=[["2001", "2010"], ["2021", "2030"]]
        )
        assert len(out.time) == 2
        assert {"2001-2010", "2021-2030"}.issubset(out.horizon.values)

    @pytest.mark.parametrize("cal", ["proleptic_gregorian", "noleap", "360_day"])
    def test_calendars(self, cal):
        ds = timeseries(
            np.tile(np.arange(1, 2), 30),
            variable="tas",
            start="2001-01-01",
            freq="YS-JAN",
            as_dataset=True,
        )

        out = xs.climatological_op(ds.convert_calendar(cal, align_on="date"), op="mean")
        assert out.time.dt.calendar == cal

    @pytest.mark.parametrize("xrfreq", ["MS", "QS-DEC", "YS-JAN"])
    def test_horizons_as_dim(self, xrfreq):
        o = 12 if xrfreq == "MS" else 4 if xrfreq == "QS-DEC" else 1
        freq = {
            "MS": {"month": o},
            "QS-DEC": {"season": o},
            "YS-JAN": {},
        }
        ds = timeseries(
            np.tile(np.arange(1, o + 1), 30),
            variable="tas",
            start="2001-01-01",
            freq=xrfreq,
            as_dataset=True,
        )

        out = xs.climatological_op(
            ds, op="mean", window=10, stride=5, horizons_as_dim=True
        )

        assert (out.tas_clim_mean.values == np.tile(np.arange(1, o + 1), (5, 1))).all()
        assert out.sizes == {"horizon": 5} | freq[xrfreq]
        assert out.time.dims == ("horizon",) + (
            (next(iter(freq[xrfreq])),) if freq[xrfreq] else ()
        )
        assert (
            out.horizon.values
            == ["2001-2010", "2006-2015", "2011-2020", "2016-2025", "2021-2030"]
        ).all()
        assert (
            out.time.values
            == np.array(
                [
                    ds.time.isel(time=slice(i, i + o)).values
                    for i in range(0, ds.time.size - 5 * o, 5 * o)
                ]
            ).squeeze()
        ).all()
        if xrfreq in ["MS", "QS-DEC"]:
            freq_coords = {
                "month": [
                    "JAN",
                    "FEB",
                    "MAR",
                    "APR",
                    "MAY",
                    "JUN",
                    "JUL",
                    "AUG",
                    "SEP",
                    "OCT",
                    "NOV",
                    "DEC",
                ],
                "season": ["MAM", "JJA", "SON", "DJF"],
                "time": np.array(
                    [[t] for t in ds.time.isel(time=range(0, 25, 5)).values]
                ),
            }
            assert (
                out[next(iter(freq.get(xrfreq)))].values
                == freq_coords[next(iter(freq.get(xrfreq)))]
            ).all()

    def test_errors(self):
        ds = timeseries(
            np.arange(1, 31),
            variable="tas",
            start="2001-01-01",
            freq="YS",
            as_dataset=True,
        )

        with pytest.raises(
            NotImplementedError,
            match="xs.climatological_op does not currently support more than one operation per call.",
        ):
            xs.climatological_op(ds, op={"mean": {}, "std": {}})

        with pytest.raises(ValueError, match="not implemented."):
            xs.climatological_op(ds, op="unknown")



import hashlib
from pathlib import Path

import numpy as np
import pytest
import xarray as xr

try:
    import xesmf as xe
except ImportError:
    xe = None
import xscen as xs
from xscen.testing import datablock_3d


class TestCreateBoundsGridmapping:
    def test_create_bounds_rotated_pole(self):
        ds = datablock_3d(
            np.zeros((20, 10, 10)),
            "tas",
            "rlon",
            -5,
            "rlat",
            80.5,
            1,
            1,
            "2000-01-01",
            as_dataset=True,
        )
        bnds = xs.regrid.create_bounds_gridmapping(ds, "rotated_pole")
        np.testing.assert_allclose(bnds.lon_bounds[-1, -1, 1], 83)
        np.testing.assert_allclose(bnds.lat_bounds[-1, -1, 1], 42.5)

        with pytest.warns(FutureWarning):
            assert xs.regrid.create_bounds_rotated_pole(ds).equals(bnds)

    def test_create_bounds_oblique(self):
        ds = datablock_3d(
            np.zeros((20, 10, 10)),
            "tas",
            "x",
            -5000,
            "y",
            5000,
            100000,
            100000,
            "2000-01-01",
            as_dataset=True,
        )
        bnds = xs.regrid.create_bounds_gridmapping(ds, "oblique_mercator")
        np.testing.assert_allclose(bnds.lon_bounds[-1, -1, -1], -48.98790806)
        np.testing.assert_allclose(bnds.lat_bounds[-1, -1, -1], 52.9169163)

    def test_error(self):
        ds = datablock_3d(
            np.zeros((20, 10, 10)),
            "tas",
            "x",
            -5000,
            "y",
            5000,
            100000,
            100000,
            "2000-01-01",
            as_dataset=True,
        )
        with pytest.raises(NotImplementedError):
            xs.regrid.create_bounds_gridmapping(ds, "lambert_conformal_conic")


@pytest.mark.skipif(xe is None, reason="xesmf needed for testing regrdding")
class TestRegridDataset:
    @staticmethod
    def compute_file_hash(file_path):
        """Compute the SHA-256 hash of the specified file."""
        sha256 = hashlib.sha256()
        with Path(file_path).open("rb") as f:
            for block in iter(lambda: f.read(4096), b""):
                sha256.update(block)
        return sha256.hexdigest()

    dsin_reg = datablock_3d(
        np.zeros((10, 6, 6)),
        "tas",
        "lon",
        -142,
        "lat",
        0,
        2,
        2,
        "2000-01-01",
        as_dataset=True,
    )
    dsin_reg = dsin_reg.chunk({"lon": 3, "time": 1})
    dsin_reg.attrs["cat:id"] = "CMIP5"
    dsin_reg.attrs["cat:member"] = "MIP5"
    dsin_reg.attrs["cat:domain"] = "Global"

    dsout_rp = datablock_3d(
        np.zeros((2, 10, 10)),
        "tas",
        "rlon",
        -5,
        "rlat",
        -5,
        1,
        1,
        "2000-01-01",
        as_dataset=True,
    )
    dsout_rp.attrs["cat:id"] = "CORDEX"
    dsout_rp.attrs["cat:domain"] = "RegionEssai"

    def test_simple(self, tmp_path):
        out = xs.regrid_dataset(
            self.dsin_reg,
            self.dsout_rp,
            weights_location=tmp_path / "weights",
            regridder_kwargs={
                "method": "patch",
                "output_chunks": {"rlon": 5},
                "unmapped_to_nan": True,
            },
        )

        assert (
            tmp_path / "weights" / "C_Global_CORDEX_RegionEssai_regrid0patch.nc"
        ).is_file()
        assert out.tas.attrs["grid_mapping"] == "rotated_pole"
        assert out.rotated_pole.attrs == self.dsout_rp.rotated_pole.attrs
        assert "patch" in out.attrs["history"]
        assert out.attrs["cat:processing_level"] == "regridded"
        assert out.chunks["rlon"] == (5, 5)

        hash1 = self.compute_file_hash(
            tmp_path / "weights" / "C_Global_CORDEX_RegionEssai_regrid0patch.nc"
        )
        xs.regrid_dataset(
            self.dsin_reg,
            self.dsout_rp,
            weights_location=tmp_path / "weights",
            regridder_kwargs={
                "method": "patch",
                "output_chunks": {"rlon": 5},
                "unmapped_to_nan": True,
            },
        )
        # Check that the weights are not recomputed
        hash2 = self.compute_file_hash(
            tmp_path / "weights" / "C_Global_CORDEX_RegionEssai_regrid0patch.nc"
        )
        assert hash1 == hash2

    def test_mask(self):
        ds_in = self.dsin_reg.copy()
        ds_in["tas"].loc[dict(lon=-142, lat=0)] = 999999
        ds_in["mask"] = xr.ones_like(ds_in.tas.isel(time=0))
        ds_in["mask"].loc[dict(lon=-142, lat=0)] = 0

        grid = xe.util.cf_grid_2d(-140, -134, 1, 2, 8, 1)
        grid["mask"] = xr.DataArray(np.ones((6, 6)), dims=("lat", "lon"))
        grid["mask"].loc[dict(lon=-134.5, lat=7.5)] = 0

        out = xs.regrid_dataset(
            ds_in,
            grid,
            regridder_kwargs={
                "method": "bilinear",
                "skipna": True,
            },
        )
        assert "mask" not in out
        np.testing.assert_equal(
            out.tas.isel(time=0), xr.where(grid["mask"] == 0, np.nan, 0).values
        )

    @pytest.mark.parametrize(
        "unmapped_to_nan, skipna",
        [[True, False], [False, True], [False, False], [None, False]],
    )
    def test_unmapped_to_nan(self, unmapped_to_nan, skipna):
        out = xs.regrid_dataset(
            self.dsin_reg,
            xe.util.cf_grid_2d(-140, -130, 1, 2, 8, 1),
            regridder_kwargs={
                "method": "bilinear",
                "skipna": skipna,
                "unmapped_to_nan": unmapped_to_nan,
            },
        )
        if skipna is False and not unmapped_to_nan:
            # This is the only case where unmapped NaNs will be extrapolated
            np.testing.assert_equal(
                out.tas.isel(time=0, lat=0), np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
            )
        else:
            np.testing.assert_equal(
                out.tas.isel(time=0, lat=0),
                np.array([0, 0, 0, 0, 0, 0, 0, 0, np.nan, np.nan]),
            )

    def test_no_regrid(self, tmp_path):
        out = xs.regrid_dataset(
            self.dsin_reg,
            self.dsin_reg,
            regridder_kwargs={
                "method": "patch",
                "output_chunks": {"rlon": 5},
                "unmapped_to_nan": True,
            },
            to_level="regridded2",
        )
        assert out.equals(self.dsin_reg)
        assert out.attrs["cat:processing_level"] == "regridded2"

        # Add a mask
        ds_in = self.dsin_reg.copy()
        ds_in["mask"] = xr.ones_like(ds_in.tas.isel(time=0))
        ds_in["mask"].loc[dict(lon=-142, lat=0)] = 0
        ds_grid = self.dsin_reg.copy()
        ds_grid["mask"] = xr.ones_like(ds_grid.tas.isel(time=0))
        ds_grid["mask"].loc[dict(lon=-142, lat=2)] = 0
        out = xs.regrid_dataset(
            ds_in,
            ds_grid,
            regridder_kwargs={
                "method": "patch",
                "output_chunks": {"rlon": 5},
                "unmapped_to_nan": True,
            },
        )
        np.testing.assert_allclose(
            out.tas.isel(time=2, lon=0), np.array([np.nan, np.nan, 0, 0, 0, 0])
        )
        np.testing.assert_allclose(
            out.tas.isel(time=2, lon=1), np.array([0, 0, 0, 0, 0, 0])
        )
        assert "mask" not in out

    def test_intermediate(self, tmp_path):
        intermediate = {
            "intermediate": {
                "cf_grid_2d": {
                    "lon0_b": -142,
                    "lon1_b": -132,
                    "d_lon": 1,
                    "lat0_b": 0,
                    "lat1_b": 10,
                    "d_lat": 1,
                },
                "regridder_kwargs": {"method": "bilinear", "skipna": True},
            }
        }
        out = xs.regrid_dataset(
            self.dsin_reg,
            self.dsout_rp,
            weights_location=tmp_path / "weights",
            intermediate_grids=intermediate,
            regridder_kwargs={
                "method": "patch",
                "output_chunks": {"rlon": 5},
                "unmapped_to_nan": True,
            },
        )
        assert (
            tmp_path / "weights" / "C_Global_CORDEX_RegionEssai_regrid0bilinear.nc"
        ).is_file()
        assert (
            tmp_path / "weights" / "C_Global_CORDEX_RegionEssai_regrid1patch.nc"
        ).is_file()
        assert "cf_grid_2d with arguments" in out.attrs["history"]

    @pytest.mark.parametrize("gridmap", ["oblique_mercator", "rotated_pole"])
    def test_conservative_in(self, tmp_path, gridmap):
        mult = 1 if gridmap == "rotated_pole" else 100000

        ds_in = datablock_3d(
            np.tile(np.arange(6), (1, 4, 1)) / 86400,
            "pr",
            "rlon" if gridmap == "rotated_pole" else "x",
            0,
            "rlat" if gridmap == "rotated_pole" else "y",
            0,
            1 * mult,
            1 * mult,
            "2000-01-01",
            as_dataset=True,
            units="kg m-2 s-1",
        )
        grid = xe.util.cf_grid_2d(
            ds_in.lon.min().values - 0.5,
            ds_in.lon.max().values + 0.5,
            0.22,
            ds_in.lat.min().values - 0.5,
            ds_in.lat.max().values + 0.5,
            0.22,
        )

        out = xs.regrid_dataset(
            ds_in,
            grid,
            regridder_kwargs={
                "method": "conservative",
                "skipna": False,
            },
        )
        assert out.attrs["regrid_method"] == "conservative"
        assert "bounds" not in out.dims
        assert "lon_bounds" not in out
        assert "lat_bounds" not in out
        assert gridmap not in out

    def test_conservative_out(self, tmp_path):
        ds_in = datablock_3d(
            np.tile(np.arange(6), (1, 4, 1)) / 86400,
            "pr",
            "lon",
            -142.5,
            "lat",
            2,
            5,
            1,
            "2000-01-01",
            as_dataset=True,
            units="kg m-2 s-1",
        )

        out = xs.regrid_dataset(
            ds_in,
            self.dsout_rp,
            regridder_kwargs={
                "method": "conservative",
                "skipna": False,
            },
        )
        assert out.attrs["regrid_method"] == "conservative"
        assert "bounds" not in out.dims
        assert "lon_bounds" not in out
        assert "lat_bounds" not in out
        assert "rotated_pole" in out.coords
        assert all(d in out.dims for d in ["rlon", "rlat"])
        assert all(c in out.coords for c in ["lon", "lat"])
        assert out.pr.attrs["grid_mapping"] == "rotated_pole"

    def test_conservative_multiple(self):
        ds_in = datablock_3d(
            np.tile(np.arange(6), (1, 4, 1)) / 86400,
            "pr",
            "lon",
            -142.5,
            "lat",
            2,
            5,
            1,
            "2000-01-01",
            as_dataset=True,
            units="kg m-2 s-1",
        )

        out = xs.regrid_dataset(
            ds_in,
            self.dsout_rp,
            regridder_kwargs={
                "method": "conservative",
                "skipna": False,
            },
        )
        assert out.attrs["regrid_method"] == "conservative"
        assert "bounds" not in out.dims
        assert "lon_bounds" not in out
        assert "lat_bounds" not in out
        assert "rotated_pole" in out.coords
        assert all(d in out.dims for d in ["rlon", "rlat"])
        assert all(c in out.coords for c in ["lon", "lat"])
        assert out.pr.attrs["grid_mapping"] == "rotated_pole"


class TestMask:
    ds = datablock_3d(
        np.tile(np.arange(6), (1, 4, 1)),
        "tas",
        "lon",
        -142.5,
        "lat",
        2,
        5,
        1,
        "2000-01-01",
        as_dataset=True,
    )
    ds["tas"] = ds["tas"].where(ds["tas"].lon > -142.5)

    @pytest.mark.parametrize("mask_nans", [True, False])
    def test_mask_simple(self, mask_nans):
        mask = xs.regrid.create_mask(
            self.ds,
            variable="tas",
            where_operator=">",
            where_threshold=2,
            mask_nans=mask_nans,
        )
        assert isinstance(mask, xr.DataArray)
        assert mask.attrs["where_threshold"] == "tas > 2"
        assert mask.attrs["mask_NaNs"] == str(mask_nans)
        assert "time" not in mask.dims
        np.testing.assert_allclose(
            mask, np.stack([np.array([0 if mask_nans else 1, 0, 0, 1, 1, 1])] * 4)
        )

        mask2 = xs.regrid.create_mask(
            self.ds["tas"], where_operator=">", where_threshold=2, mask_nans=mask_nans
        )
        assert mask2.equals(mask)

    def test_units(self):
        mask = xs.regrid.create_mask(
            self.ds, variable="tas", where_operator=">=", where_threshold="2 K"
        )
        assert mask.attrs["where_threshold"] == "tas >= 2 K"
        np.testing.assert_allclose(mask, np.stack([np.array([0, 0, 1, 1, 1, 1])] * 4))

        mask2 = xs.regrid.create_mask(
            self.ds, variable="tas", where_operator=">=", where_threshold="2 C"
        )
        assert mask2.attrs["where_threshold"] == "tas >= 2 C"
        np.testing.assert_allclose(mask2, np.stack([np.array([0, 0, 0, 0, 0, 0])] * 4))

    def test_error(self):
        with pytest.raises(
            ValueError,
            match="'where_operator' and 'where_threshold' must be used together.",
        ):
            xs.regrid.create_mask(self.ds, variable="tas", where_operator=">")
        with pytest.raises(
            ValueError, match="A variable needs to be specified when passing a Dataset."
        ):
            xs.regrid.create_mask(self.ds)



"""Functions to train and adjust a dataset using a bias-adjustment algorithm."""

import logging
import warnings
from copy import deepcopy

import xarray as xr
import xclim as xc
import xsdba
from xclim.core.units import infer_context

from .catutils import parse_from_ds
from .config import parse_config
from .utils import minimum_calendar, standardize_periods

logger = logging.getLogger(__name__)
xsdba.set_options(extra_output=False)


__all__ = [
    "adjust",
    "train",
]


def _add_preprocessing_attr(scen, train_kwargs):
    fake_ref = xr.DataArray(name="ref")
    fake_hist = xr.DataArray(name="hist")

    preproc = []
    if train_kwargs["jitter_under"] is not None:
        preproc.append(
            xc.core.formatting.gen_call_string(
                "jitter_under_thresh", fake_ref, fake_hist, train_kwargs["jitter_under"]
            )
        )
    if train_kwargs["jitter_over"] is not None:
        preproc.append(
            xc.core.formatting.gen_call_string(
                "jitter_over_thresh", fake_ref, fake_hist, train_kwargs["jitter_over"]
            )
        )
    if train_kwargs["adapt_freq"] is not None:
        preproc.append(
            xc.core.formatting.gen_call_string(
                "adapt_freq", fake_ref, fake_hist, train_kwargs["adapt_freq"]
            )
        )

    if preproc:
        scen.attrs[
            "bias_adjustment"
        ] += ", ref and hist were prepared with " + " and ".join(preproc)
    return scen


@parse_config
def train(
    dref: xr.Dataset,
    dhist: xr.Dataset,
    var: str | list[str],
    period: list[str],
    *,
    method: str = "DetrendedQuantileMapping",
    group: xsdba.Grouper | str | dict | None = None,
    xsdba_train_args: dict | None = None,
    xclim_train_args: dict | None = None,
    maximal_calendar: str = "noleap",
    adapt_freq: dict | None = None,
    jitter_under: dict | None = None,
    jitter_over: dict | None = None,
    align_on: str | None = "year",
) -> xr.Dataset:
    """
    Train a bias-adjustment.

    Parameters
    ----------
    dref : xr.Dataset
      The target timeseries, on the reference period.
    dhist : xr.Dataset
      The timeseries to adjust, on the reference period.
    var : str or list of str
      Variable on which to do the adjustment. Currently only supports one variable.
    period : list of str
      [start, end] of the reference period
    method : str
      Name of the `xsdba.TrainAdjust` method of xclim.
    group : str or xsdba.Grouper or dict, optional
      Grouping information. If a string, it is interpreted as a grouper on the time dimension. If a dict, it is passed to `xsdba.Grouper.from_kwargs`.
      Defaults to {"group": "time.dayofyear", "window": 31}.
    xsdba_train_args : dict, optional
      Dict of arguments to pass to the `.train` of the adjustment object.
    xclim_train_args : dict, optional
      Dict of arguments to pass to the `.train` of the adjustment object.
      A warning will be emitted stating that this a legacy argument replaced with `xsdba_train_args`.
    maximal_calendar: str
      Maximal calendar dhist can be. The hierarchy: 360_day < noleap < standard < all_leap.
      If dhist's calendar is higher than maximal calendar, it will be converted to the maximal calendar.
    adapt_freq: dict, optional
      If given, a dictionary of args to pass to the frequency adaptation function.
    jitter_under: dict, optional
      If given, a dictionary of args to pass to `jitter_under_thresh`.
    jitter_over: dict, optional
      If given, a dictionary of args to pass to `jitter_over_thresh`.
    align_on: str, optional
      `align_on` argument for the function `xr.DataArray.convert_calendar`.

    Returns
    -------
    xr.Dataset
      Trained algorithm's data.

    See Also
    --------
    xsdba.adjustment.DetrendedQuantileMapping, xsdba.adjustment.ExtremeValues

    """
    if xclim_train_args is not None:
        warnings.warn(
            "`xclim_train_args` will be deprecated and replaced by `xsdba_train_args`.",
            FutureWarning,
        )
        if xsdba_train_args is not None:
            warnings.warn(
                "`xclim_train_args` and `xsdba_train_args` were both given, but correspond to the same option. `xsdba_train_args` will be kept"
            )
        else:
            xsdba_train_args = deepcopy(xclim_train_args)
    # TODO: To be adequately fixed later when we add multivariate
    if isinstance(var, str):
        var = [var]
    if len(var) != 1:
        raise ValueError(
            "biasadjust currently does not support entries with multiple variables."
        )
    else:
        ref = dref[var[0]]
        hist = dhist[var[0]]

    group = group if group is not None else {"group": "time.dayofyear", "window": 31}

    xsdba_train_args = xsdba_train_args or {}
    if method == "DetrendedQuantileMapping":
        xsdba_train_args.setdefault("nquantiles", 15)

    # cut out the right period
    period = standardize_periods(period, multiple=False)
    hist = hist.sel(time=slice(period[0], period[1]))
    ref = ref.sel(time=slice(period[0], period[1]))

    # convert calendar if necessary
    simcal = hist.time.dt.calendar
    refcal = ref.time.dt.calendar
    mincal = minimum_calendar(simcal, maximal_calendar)
    if simcal != mincal:
        hist = hist.convert_calendar(mincal, align_on=align_on)
    if refcal != mincal:
        ref = ref.convert_calendar(mincal, align_on=align_on)

    if isinstance(group, dict):
        # So we can specify window and add_dims in yaml.
        group = xsdba.Grouper.from_kwargs(**group)["group"]
    elif isinstance(group, str):
        group = xsdba.Grouper(group)
    xsdba_train_args["group"] = group

    # TODO: change this to be compatible with multivar too?
    contexts = [
        infer_context(da.attrs.get("standard_name", None)) for da in [ref, hist]
    ]
    cntx = "hydro" if "hydro" in contexts else "none"
    with xc.core.units.units.context(cntx):
        if jitter_over is not None:
            ref = xsdba.processing.jitter_over_thresh(ref, **jitter_over)
            hist = xsdba.processing.jitter_over_thresh(hist, **jitter_over)

        if jitter_under is not None:
            ref = xsdba.processing.jitter_under_thresh(ref, **jitter_under)
            hist = xsdba.processing.jitter_under_thresh(hist, **jitter_under)

        if adapt_freq is not None:
            adapt_freq.setdefault("group", group)
            hist, pth, dP0 = xsdba.processing.adapt_freq(ref, hist, **adapt_freq)
            adapt_freq.pop("group")

        ADJ = getattr(xsdba.adjustment, method).train(ref, hist, **xsdba_train_args)

    if adapt_freq is not None:
        ds = ADJ.ds.assign(pth=pth, dP0=dP0)
    else:
        ds = ADJ.ds

    # Arguments that need to be transferred to the adjust() function
    ds.attrs["train_params"] = {
        "var": var,
        "maximal_calendar": maximal_calendar,
        "adapt_freq": adapt_freq,
        "jitter_under": jitter_under,
        "jitter_over": jitter_over,
    }

    # attrs that are needed to open with .to_dataset_dict()
    for a in ["cat:xrfreq", "cat:domain", "cat:id"]:
        ds.attrs[a] = dhist.attrs[a] if a in dhist.attrs else None
    ds.attrs["cat:processing_level"] = f"training_{var[0]}"

    return ds


@parse_config
def adjust(
    dtrain: xr.Dataset,
    dsim: xr.Dataset,
    periods: list[str] | list[list[str]],
    *,
    xsdba_adjust_args: dict | None = None,
    xclim_adjust_args: dict | None = None,
    to_level: str = "biasadjusted",
    bias_adjust_institution: str | None = None,
    bias_adjust_project: str | None = None,
    align_on: str | None = "year",
) -> xr.Dataset:
    """
    Adjust a simulation.

    Parameters
    ----------
    dtrain : xr.Dataset
      A trained algorithm's dataset, as returned by `train`.
    dsim : xr.Dataset
      Simulated timeseries, projected period.
    periods : list of str or list of lists of str
      Either [start, end] or list of [start, end] of the simulation periods to be adjusted (one at a time).
    xsdba_adjust_args : dict, optional
      Dict of arguments to pass to the `.adjust` of the adjustment object.
    xclim_adjust_args : dict, optional
      Dict of arguments to pass to the `.adjust` of the adjustment object
      A warning will be emitted stating that this a legacy argument replaced with `xclim_train_args`.
    to_level : str
      The processing level to assign to the output.
      Defaults to 'biasadjusted'
    bias_adjust_institution : str, optional
      The institution to assign to the output.
    bias_adjust_project : str, optional
      The project to assign to the output.
    align_on: str, optional
      `align_on` argument for the function `xr.DataArray.convert_calendar`.

    Returns
    -------
    xr.Dataset
      dscen, the bias-adjusted timeseries.

    See Also
    --------
    xsdba.adjustment.DetrendedQuantileMapping, xsdba.adjustment.ExtremeValues

    """
    if xclim_adjust_args is not None:
        warnings.warn(
            "`xclim_adjust_args` will be deprecated and replaced by `xsdba_adjust_args`.",
            FutureWarning,
        )
        if xsdba_adjust_args is not None:
            warnings.warn(
                "`xclim_adjust_args` and `xsdba_adjust_args` were both given, but correspond to the same option. `xsdba_adjust_args` will be kept"
            )
        else:
            xsdba_adjust_args = deepcopy(xclim_adjust_args)
    xsdba_adjust_args = deepcopy(xsdba_adjust_args)
    xsdba_adjust_args = xsdba_adjust_args or {}

    # evaluate the dict that was stored as a string
    if not isinstance(dtrain.attrs["train_params"], dict):
        # FIXME: eval is bad. There has to be a better way!™
        dtrain.attrs["train_params"] = eval(dtrain.attrs["train_params"])  # noqa: S307

    var = dtrain.attrs["train_params"]["var"]
    if len(var) != 1:
        raise ValueError(
            "biasadjust currently does not support entries with multiple variables."
        )
    else:
        var = var[0]
        sim = dsim[var]

    # get right calendar
    simcal = sim.time.dt.calendar
    mincal = minimum_calendar(simcal, dtrain.attrs["train_params"]["maximal_calendar"])
    if simcal != mincal:
        sim = sim.convert_calendar(mincal, align_on=align_on)

    # adjust
    ADJ = xsdba.adjustment.TrainAdjust.from_dataset(dtrain)

    if ("detrend" in xsdba_adjust_args) and (
        isinstance(xsdba_adjust_args["detrend"], dict)
    ):
        name, kwargs = list(xsdba_adjust_args["detrend"].items())[0]
        kwargs = kwargs or {}
        kwargs.setdefault("group", ADJ.group)
        kwargs.setdefault("kind", ADJ.kind)
        xsdba_adjust_args["detrend"] = getattr(xsdba.detrending, name)(**kwargs)

    cntx = infer_context(sim.attrs.get("standard_name", None))
    with xc.core.units.units.context(cntx):
        # do the adjustment for all the simulation_period lists
        periods = standardize_periods(periods)
        slices = []
        for period in periods:
            sim_sel = sim.sel(time=slice(period[0], period[1]))

            out = ADJ.adjust(sim_sel, **xsdba_adjust_args)
            slices.extend([out])
        # put all the adjusted period back together
    dscen = xr.concat(slices, dim="time")

    dscen = _add_preprocessing_attr(dscen, dtrain.attrs["train_params"])
    dscen = xr.Dataset(data_vars={var: dscen}, attrs=dsim.attrs)
    dscen.attrs["cat:processing_level"] = to_level
    dscen.attrs["cat:variable"] = parse_from_ds(dscen, ["variable"])["variable"]
    if bias_adjust_institution is not None:
        dscen.attrs["cat:bias_adjust_institution"] = bias_adjust_institution
    if bias_adjust_project is not None:
        dscen.attrs["cat:bias_adjust_project"] = bias_adjust_project

    return dscen



"""Functions to aggregate data over time and space."""

import datetime
import logging
import os
import warnings
from collections.abc import Sequence
from copy import deepcopy
from pathlib import Path
from types import ModuleType

import geopandas as gpd
import numpy as np
import pandas as pd
import scipy
import shapely
import xarray as xr
import xclim as xc
import xclim.core.calendar

try:
    import xesmf as xe
except ImportError:
    xe = None
from xclim.core.indicator import Indicator
from xclim.core.units import pint2cfattrs, units2pint

from .config import parse_config
from .extract import subset_warming_level
from .indicators import compute_indicators
from .spatial import get_grid_mapping, subset
from .utils import standardize_periods, unstack_dates, update_attr

logger = logging.getLogger(__name__)

__all__ = [
    "climatological_op",
    "compute_deltas",
    "produce_horizon",
    "spatial_mean",
]


# Dummy function to make gettext aware of translatable-strings
def _(s):
    return s


@parse_config
def climatological_op(  # noqa: C901
    ds: xr.Dataset,
    *,
    op: str | dict = "mean",
    window: int | None = None,
    min_periods: int | float | None = None,
    stride: int = 1,
    periods: list[str] | list[list[str]] | None = None,
    rename_variables: bool = True,
    to_level: str = "climatology",
    horizons_as_dim: bool = False,
) -> xr.Dataset:
    """
    Perform an operation 'op' over time, for given time periods, respecting the temporal resolution of ds.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to use for the computation.
    op : str or dict
        Operation to perform over time.
        The operation can be any method name of xarray.core.rolling.DatasetRolling, 'linregress', or a dictionary.
        If 'op' is a dictionary, the key is the operation name and the value is a dict of kwargs
        accepted by the equivalent NumPy function. See the Notes for more information.
        While other operations are technically possible, the following are recommended and tested:
        ['max', 'mean', 'median', 'min', 'std', 'sum', 'var', 'linregress'].
        Operations beyond methods of xarray.core.rolling.DatasetRolling include:

            - 'linregress' : Computes the linear regression over time, using
              scipy.stats.linregress and employing years as regressors.
              The output will have a new dimension 'linreg_param' with coordinates:
              ['slope', 'intercept', 'rvalue', 'pvalue', 'stderr', 'intercept_stderr'].

        Only one operation per call is supported, so len(op)==1 if a dict.
    window : int, optional
        Number of years to use for the rolling operation.
        If left at None and periods is given, window will be the size of the first period. Hence, if periods are of
        different lengths, the shortest period should be passed first.
        If left at None and periods is not given, the window will be the size of the input dataset.
    min_periods : int or float, optional
        For the rolling operation, minimum number of years required for a value to be computed.
        If left at None and the xrfreq is either QS or AS and doesn't start in January,
        min_periods will be one less than window.
        Otherwise, if left at None, it will be deemed the same as 'window'.
        If passed as a float value between 0 and 1, this will be interpreted as the floor of the fraction of the window size.
    stride : int
        Stride (in years) at which to provide an output from the rolling window operation.
    periods : list of str or list of lists of str, optional
        Either [start, end] or list of [start, end] of continuous periods to be considered.
        This is needed when the time axis of ds contains some jumps in time.
        If None, the dataset will be considered continuous.
    rename_variables : bool
        If True, '_clim_{op}' will be added to variable names.
    to_level : str, optional
        The processing level to assign to the output.
        If None, the processing level of the inputs is preserved.
    horizons_as_dim : bool
        If True, the output will have 'horizon' and the frequency as 'month', 'season' or 'year' as
        dimensions and coordinates. The 'time' coordinate will be unstacked to horizon and frequency dimensions.
        Horizons originate from periods and/or windows and their stride in the rolling operation.

    Returns
    -------
    xr.Dataset
        Dataset with the results from the climatological operation.

    Notes
    -----
    xarray.core.rolling.DatasetRolling functions do not support kwargs other than 'keep_attrs'. In order to pass
    additional arguments to the operation, we instead use the 'reduce' method and pass the operation as a numpy function.
    If possible, a function that handles NaN values will be used (e.g. op='mean' will use `np.nanmean`), as the
    'min_periods' argument already decides how many NaN values are acceptable.
    """
    # Daily data is not supported
    if len(ds.time) > 3 and xr.infer_freq(ds.time) == "D":
        raise NotImplementedError(
            "xs.climatological_op does not currently support daily data."
        )
    # more than one operation per call is not supported (yet), case for dict
    if isinstance(op, dict) and len(op) > 1:
        raise NotImplementedError(
            "xs.climatological_op does not currently support more than one operation per call."
        )

    # unstack 1D time in coords (day, month, and year) to make climatological mean faster
    mindex_coords = xr.Coordinates.from_pandas_multiindex(
        pd.MultiIndex.from_arrays(
            [
                ds.time.dt.year.values,
                ds.time.dt.month.values,
                ds.time.dt.day.values,
            ],
            names=["year", "month", "day"],
        ),
        dim="time",
    )
    ds_unstack = ds.assign_coords(coords=mindex_coords).unstack("time")

    # Rolling will ignore gaps in time, so raise an exception beforehand
    if (not all(ds_unstack.year.diff(dim="year", n=1) == 1)) & (periods is None):
        raise ValueError("Data is not continuous. Use the 'periods' argument.")

    # define periods, windows, and min_periods
    periods = standardize_periods(
        periods or [[int(ds_unstack.year[0]), int(ds_unstack.year[-1])]]
    )
    window = window or int(periods[0][1]) - int(periods[0][0]) + 1

    # there is one less occurrence when a period crosses years
    freq_across_year = [
        f"{f}-{mon}"
        for mon in xr.coding.cftime_offsets._MONTH_ABBREVIATIONS.values()
        for f in ["AS", "QS"]
        if mon != "JAN"
    ]
    if (
        any(
            x in freq_across_year
            for x in [
                ds.attrs.get("cat:xrfreq"),
                (xr.infer_freq(ds.time) if len(ds.time) > 3 else None),
            ]
        )
        and min_periods is None
    ):
        min_periods = window - 1

    # unpack min_periods as fraction of window
    if isinstance(min_periods, float):
        if 0 < min_periods <= 1:
            min_periods = int(np.floor(min_periods * window))
        else:
            raise ValueError(
                f"When 'min_periods' is passed as a 'float', it must be between 0 and 1. Got {min_periods}."
            )

    # set min_periods
    min_periods = min_periods or window
    if min_periods > window:
        raise ValueError("'min_periods' should be smaller or equal to 'window'")

    # if op is a dict, unpack it
    if isinstance(op, dict):
        op, op_kwargs = list(op.items())[0]
        op_kwargs.setdefault("keep_attrs", True)
    else:
        op_kwargs = {"keep_attrs": True}

    # special case for averaging standard deviations: need to convert to variance before averaging
    ds_has_std = False
    std_v = []
    if op == "mean":
        for vv in ds_unstack.data_vars:
            if (
                "std" in vv
                or "standard deviation"
                in ds_unstack[vv].attrs.get("description", "").lower()
            ):
                ds_unstack[vv] = np.square(ds_unstack[vv])
                ds_has_std = True
                std_v.extend([vv])

    # Compute the climatological operation
    concats = []
    for period in periods:
        # Rolling average
        ds_rolling = ds_unstack.sel(year=slice(period[0], period[1])).rolling(
            year=window, min_periods=min_periods
        )

        # apply operation on rolling object
        if hasattr(ds_rolling, op) and callable(getattr(ds_rolling, op)):
            if op not in ["max", "mean", "median", "min", "std", "sum", "var"]:
                warnings.warn(
                    f"The requested operation '{op}' has not been tested and may produce unexpected results."
                )
            if len(set(op_kwargs.keys()).difference(["keep_attrs"])) > 0:
                # Reduce allows for kwargs to be passed to the operation, but it is less efficient
                try:
                    numpy_op = getattr(np, f"nan{op}")
                except AttributeError:
                    warnings.warn(
                        f"The requested operation '{op}' has no NaN handling in numpy. Results may ignore the 'min_periods' argument."
                    )
                    numpy_op = getattr(np, op)
                ds_rolling = ds_rolling.reduce(numpy_op, **op_kwargs)
            else:
                ds_rolling = getattr(ds_rolling, op)(**op_kwargs)

            # revert variance to std, where applicable
            if op == "mean" and ds_has_std:
                for vv in std_v:
                    ds_rolling[vv] = np.sqrt(ds_rolling[vv])

            # Select the windows at provided stride, starting from the first full window's operation result
            ds_rolling = ds_rolling.isel(year=slice(window - 1, None, stride))

        elif op == "linregress":

            def _ulinregress(x, y, **kwargs):
                # Wrapper for scipy.stats.linregress to unpack multiple return values in xr.apply_ufunc
                valid_x = ~np.isnan(x)
                valid_y = ~np.isnan(y)
                mask = valid_x & valid_y
                if np.sum(mask) >= kwargs.get("min_periods", 1):
                    x = x[mask]
                    y = y[mask]
                    reg = scipy.stats.linregress(
                        x, y, alternative=kwargs.get("alternative", "two-sided")
                    )
                    out = np.array(
                        [
                            reg.slope,
                            reg.intercept,
                            reg.rvalue,
                            reg.pvalue,
                            reg.stderr,
                            reg.intercept_stderr,
                        ]
                    )
                else:
                    out = np.full(6, np.nan)
                return out

            # prepare kwargs
            linreg_kwargs = {
                k: v for k, v in op_kwargs.items() if "keep_attrs" not in k
            }
            linreg_kwargs["min_periods"] = min_periods

            # unwrap DatasetRolling object and select years subset
            dsr_construct = ds_rolling.construct(window_dim="window", keep_attrs=True)
            dsr_construct = dsr_construct.isel(year=slice(window - 1, None, stride))

            # construct array to use years as x values (==regressors) in xr.apply_ufunc
            years_as_x_values = xr.DataArray(
                np.arange(dsr_construct.window.size)
                .repeat(dsr_construct.year.size)
                .reshape(dsr_construct.window.size, dsr_construct.year.size)
                + dsr_construct.year.values
                - window
                + 1,
                dims=["window", "year"],
                coords={
                    "window": dsr_construct.window.values,
                    "year": dsr_construct.year.values,
                },
            )

            # apply linregress along windows
            ds_rolling = xr.apply_ufunc(
                _ulinregress,
                years_as_x_values,
                dsr_construct,
                input_core_dims=[["window"], ["window"]],
                output_core_dims=[["linreg_param"]],
                vectorize=True,
                dask="parallelized",
                output_dtypes=["float32"],
                dask_gufunc_kwargs={"output_sizes": {"linreg_param": 6}},
                keep_attrs="no_conflicts",
                kwargs=linreg_kwargs,
            )
            # label new coords
            ds_rolling.coords["linreg_param"] = [
                "slope",
                "intercept",
                "rvalue",
                "pvalue",
                "stderr",
                "intercept_stderr",
            ]

        else:
            raise ValueError(f"Operation '{op}' not implemented.")

        # build horizons
        horizons = xr.DataArray(
            [f"{yr - (window - 1)}-{yr}" for yr in ds_rolling.year.values],
            dims=dict(year=ds_rolling.year),
        ).astype(str)
        ds_rolling = ds_rolling.assign_coords(horizon=horizons)

        # revert to 1D time, rebuilding time coord
        ds_rolling = ds_rolling.stack(time=("year", "month", "day"))
        if isinstance(ds.indexes["time"], pd.core.indexes.datetimes.DatetimeIndex):
            time_coord = pd.to_datetime(
                {
                    "year": ds_rolling.year.values - window + 1,
                    "month": ds_rolling.month.values,
                    "day": ds_rolling.day.values,
                }
            ).to_list()
        elif isinstance(ds.indexes["time"], xr.coding.cftimeindex.CFTimeIndex):
            time_coord = [
                xclim.core.calendar.datetime_classes[ds.time.dt.calendar](
                    y - window + 1, m, d
                )
                for y, m, d in zip(
                    ds_rolling.year.values,
                    ds_rolling.month.values,
                    ds_rolling.day.values,
                )
            ]
        else:
            raise ValueError("The type of 'time' was not understood.")
        ds_rolling = ds_rolling.drop_vars({"month", "year", "time", "day"})
        ds_rolling = ds_rolling.assign_coords(time=time_coord).transpose("time", ...)

        # append to list of results
        concats.extend([ds_rolling])
        # end loop over periods

    # concatenate results
    ds_rolling = xr.concat(concats, dim="time", data_vars="minimal")

    # update data_vars names, attrs, history
    if rename_variables:
        ds_rolling = ds_rolling.rename_vars(
            {vv: f"{vv}_clim_{op}" for vv in ds_rolling.data_vars}
        )

    for vv in ds_rolling.data_vars:
        for a in ["description", "long_name"]:
            try:
                op_format = dict.fromkeys(
                    ("mean", "std", "var", "sum"), "adj"
                ) | dict.fromkeys(("max", "min"), "noun")
                operation = xc.core.formatting.default_formatter.format_field(
                    op, op_format[op]
                )
            except (KeyError, ValueError):
                operation = op
            update_attr(
                ds_rolling[vv],
                a,
                _("{window}-year climatological {operation} of {attr}."),
                window=window,
                operation=operation,
            )

        new_history = (
            f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] {window}-year climatological {operation} "
            f"over window (non-centered), with a minimum of {min_periods} years of data - xarray v{xr.__version__}"
        )
        history = (
            new_history + " \n " + ds_rolling[vv].attrs["history"]
            if "history" in ds_rolling[vv].attrs
            else new_history
        )
        ds_rolling[vv].attrs["history"] = history

    # update processing level
    if to_level is not None:
        ds_rolling.attrs["cat:processing_level"] = to_level

    if horizons_as_dim:
        # restructure output to have horizons as a dimension instead of stacked horizons per year/season/month
        new_coords = {1: "year", 4: "season", 12: "month"}
        new_coord_len = len(np.unique(ds_rolling.time.dt.month))

        if new_coord_len == 1:
            all_horizons = [
                ds_rolling.sel(time=ds_rolling.horizon == horizon)
                .swap_dims({"time": "horizon"})
                .assign(
                    time_2D=xr.DataArray(
                        ds_rolling.time.sel(
                            time=ds_rolling.horizon == horizon
                        ).values.copy(),
                        dims=["time"],
                    )
                )
                .drop_vars("time")
                .rename({"time_2D": "time"})
                .set_coords("time")
                .squeeze(dim="time")
                for horizon in np.unique(ds_rolling.horizon.values)
            ]
        else:
            all_horizons = [
                unstack_dates(
                    ds_rolling.sel(time=ds_rolling.horizon == horizon).assign(
                        time_2D=xr.DataArray(
                            ds_rolling.time.sel(
                                time=ds_rolling.horizon == horizon
                            ).values.copy(),
                            dims=["time"],
                        )
                    ),
                    new_dim=new_coords[new_coord_len],
                )
                .drop_vars("horizon")
                .assign_coords(horizon=("time", [horizon]))
                .swap_dims({"time": "horizon"})
                .drop_vars("time")
                .rename({"time_2D": "time"})
                .set_coords("time")
                for horizon in np.unique(ds_rolling.horizon.values)
            ]

        return xr.concat(all_horizons, dim="horizon")

    else:
        return ds_rolling


@parse_config
def compute_deltas(  # noqa: C901
    ds: xr.Dataset,
    reference_horizon: str | xr.Dataset,
    *,
    kind: str | dict = "+",
    rename_variables: bool = True,
    to_level: str | None = "deltas",
) -> xr.Dataset:
    """Compute deltas in comparison to a reference time period, respecting the temporal resolution of ds.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to use for the computation.
    reference_horizon : str or xr.Dataset
        Either a YYYY-YYYY string corresponding to the 'horizon' coordinate of the reference period,
        or a xr.Dataset containing the climatological mean.
    kind : str or dict
        ['+', '/', '%'] Whether to provide absolute, relative, or percentage deltas.
        Can also be a dictionary separated per variable name.
    rename_variables : bool
        If True, '_delta_YYYY-YYYY' will be added to variable names.
    to_level : str, optional
        The processing level to assign to the output.
        If None, the processing level of the inputs is preserved.

    Returns
    -------
    xr.Dataset
        Returns a Dataset with the requested deltas.
    """
    if isinstance(reference_horizon, str):
        if reference_horizon not in ds.horizon:
            raise ValueError(
                f"The reference horizon {reference_horizon} is not in the dataset."
            )
        # Separate the reference from the other horizons
        if xc.core.utils.uses_dask(ds["horizon"]):
            ds["horizon"].load()
        ref = ds.where(ds.horizon == reference_horizon, drop=True)
    elif isinstance(reference_horizon, xr.Dataset):
        ref = reference_horizon
        if "horizon" in ref:
            reference_horizon = np.unique(ref["horizon"])
            if len(reference_horizon) != 1:
                raise ValueError(
                    "The reference dataset appears to contain multiple horizons."
                )
            reference_horizon = reference_horizon[0]
        else:
            reference_horizon = "unknown_horizon"
    else:
        raise ValueError(
            "reference_horizon should be either a string or an xarray.Dataset."
        )

    if "time" in ds:
        if (len(ds.time) >= 3) and (xr.infer_freq(ds.time) == "D"):
            raise NotImplementedError(
                "xs.compute_deltas does not currently support daily data."
            )

        # Remove references to 'year' in REF
        mindex_coords_1 = xr.Coordinates.from_pandas_multiindex(
            pd.MultiIndex.from_arrays(
                [ref.time.dt.month.values, ref.time.dt.day.values],
                names=["month", "day"],
            ),
            dim="time",
        )
        ref = ref.assign_coords(coords=mindex_coords_1).unstack("time")

        mindex_coords_2 = xr.Coordinates.from_pandas_multiindex(
            pd.MultiIndex.from_arrays(
                [
                    ds.time.dt.year.values,
                    ds.time.dt.month.values,
                    ds.time.dt.day.values,
                ],
                names=["year", "month", "day"],
            ),
            dim="time",
        )
        other_hz = ds.assign_coords(coords=mindex_coords_2).unstack("time")

    else:
        other_hz = ds
        ref = ref.squeeze()
    deltas = xr.Dataset(coords=other_hz.coords, attrs=other_hz.attrs)
    # Calculate deltas
    for vv in list(ds.data_vars):
        v_name = (
            vv
            if rename_variables is False
            else f"{vv}_delta_{reference_horizon.replace('-', '_')}"
        )

        with xr.set_options(keep_attrs=True):
            if (isinstance(kind, dict) and kind[vv] == "+") or kind == "+":
                _kind = "abs."
                deltas[v_name] = other_hz[vv] - ref[vv]
                unit = pint2cfattrs(
                    units2pint(other_hz[vv].attrs["units"]), is_difference=True
                )
                deltas[v_name].attrs.update(unit)
            elif (isinstance(kind, dict) and kind[vv] == "/") or kind == "/":
                _kind = "rel."
                deltas[v_name] = other_hz[vv] / ref[vv]
                deltas[v_name].attrs["units"] = ""
            elif (isinstance(kind, dict) and kind[vv] == "%") or kind == "%":
                _kind = "pct."
                deltas[v_name] = 100 * (other_hz[vv] - ref[vv]) / ref[vv]
                deltas[v_name].attrs["units"] = "%"
            else:
                raise ValueError(
                    f"Delta 'kind' not understood. Should be '+', '/' or '%', received {kind}."
                )

        # modify attrs and history
        deltas[v_name].attrs["delta_kind"] = _kind
        deltas[v_name].attrs["delta_reference"] = reference_horizon

        for a in ["description", "long_name"]:
            update_attr(
                deltas[v_name],
                a,
                _("{attr1}: {kind} delta compared to {refhoriz}."),
                others=[other_hz[vv]],
                refhoriz=reference_horizon,
                kind=_kind,
            )

        new_history = f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] {_kind} delta vs. {reference_horizon} - xarray v{xr.__version__}"
        history = (
            new_history + " \n " + deltas[v_name].attrs["history"]
            if "history" in deltas[v_name].attrs
            else new_history
        )
        deltas[v_name].attrs["history"] = history

    if "time" in ds:
        # get back to 1D time
        deltas = deltas.stack(time=("year", "month", "day"))
        # rebuild time coord
        if isinstance(ds.indexes["time"], pd.core.indexes.datetimes.DatetimeIndex):
            time_coord = list(
                pd.to_datetime(
                    {
                        "year": deltas.year.values,
                        "month": deltas.month.values,
                        "day": deltas.day.values,
                    }
                ).values
            )
        elif isinstance(ds.indexes["time"], xr.coding.cftimeindex.CFTimeIndex):
            time_coord = [
                xclim.core.calendar.datetime_classes[ds.time.dt.calendar](y, m, d)
                for y, m, d in zip(
                    deltas.year.values, deltas.month.values, deltas.day.values
                )
            ]
        else:
            raise ValueError("The type of 'time' could not be understood.")

        deltas = (
            deltas.drop_vars({"year", "day", "month", "time"})
            .assign(time=time_coord)
            .transpose("time", ...)
        )
        deltas = deltas.reindex_like(ds)

    if to_level is not None:
        deltas.attrs["cat:processing_level"] = to_level

    return deltas


@parse_config
def spatial_mean(  # noqa: C901
    ds: xr.Dataset,
    method: str,
    *,
    spatial_subset: bool | None = None,
    region: dict | str | None = None,
    simplify_tolerance: float | None = None,
    kwargs: dict | None = None,
    to_domain: str | None = None,
    to_level: str | None = None,
) -> xr.Dataset:
    r"""Compute the spatial mean using a variety of available methods.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to use for the computation.
    method : str
        'cos-lat' will weight the area covered by each pixel using an approximation based on latitude.
        'xesmf' will make use of xESMF's SpatialAverager. This will typically be more precise,
        especially for irregular regions, but can be much slower than other methods.
    spatial_subset : bool, optional
        If True, xscen.spatial.subset will be called prior to the other operations. This requires the 'region' argument.
        If None, this will automatically become True if 'region' is provided and the subsetting method is 'cos-lat'.
    region : dict or str, optional
        Description of the region and the subsetting method (required fields listed in the Notes).
        If method=='xesmf', the bounding box or shapefile is given to SpatialAverager.
        Can also be "global", for global averages.
        The latter is simply a shortcut for `{'name': 'global', 'method': 'bbox', 'lon_bnds' [-180, 180], 'lat_bnds': [-90, 90]}`.
    simplify_tolerance : float, optional
        Only used with 'xesmf' method. Precision (in degree) used to simplify a shapefile before sending it to SpatialAverager().
        The simpler the polygons, the faster the averaging, but it will lose some precision.
    kwargs : dict, optional
        Arguments to send to SpatialAverager().
        In addition, also for SpatialAverager, one can give `skipna` or  `output_chunks` here, to be passed to the averager call itself.
        Additionally, for all methods, the 'latitude_name' and 'longitude_name' can be passed here to specify the names of those coordinates.
    to_domain : str, optional
        The domain to assign to the output.
        If None, the domain of the inputs is preserved.
    to_level : str, optional
        The processing level to assign to the output.
        If None, the processing level of the inputs is preserved.

    Returns
    -------
    xr.Dataset
        Returns a Dataset with the spatial dimensions averaged.

    Notes
    -----
    'region' required fields:
        name: str
            Region name used to overwrite domain in the catalog.
        method: str
            ['gridpoint', 'bbox', shape', 'sel']
        tile_buffer: float, optional
            Multiplier to apply to the model resolution. Only used if spatial_subset==True.
        \*\*kwargs
            Arguments specific to the method used.

    See Also
    --------
    xesmf.SpatialAverager
    """
    kwargs = (kwargs or {}).copy()

    # Determine the coordinates
    i = 0
    while (
        "latitude" not in ds.cf.coordinates or "longitude" not in ds.cf.coordinates
    ) and i < 2:
        if i == 0:
            # First, try to guess the latitude coordinate using CF conventions
            ds = ds.cf.guess_coord_axis()
        elif i == 1:
            missing_coords = [
                f"{coord}_name"
                for coord in ["latitude", "longitude"]
                if coord not in ds.cf.coordinates
            ]
            # If the coordinates are still not found, use the 'kwargs' argument
            if any(coord not in kwargs for coord in missing_coords):
                raise ValueError(
                    "Could not determine the spatial coordinate names using CF conventions. "
                    "Use kwargs = {latitude_name: str, longitude_name: str} to specify them."
                )
            for coord in missing_coords:
                ds[kwargs[coord]].attrs["standard_name"] = coord.split("_")[0]
        i += 1
    kwargs.pop("latitude_name", None)
    kwargs.pop("longitude_name", None)

    if region == "global":
        region = {
            "name": "global",
            "method": "bbox",
            "lat_bnds": [-90 + 1e-5, 90 - 1e-5],
        }
        # `spatial_subset` won't wrap coords on the bbox, we need to fit the system used on ds.
        if ds.cf["longitude"].min() >= 0:
            region["lon_bnds"] = [0, 360]
        else:
            region["lon_bnds"] = [-180, 180]

    if (region is not None) and (spatial_subset is None) and (method == "cos-lat"):
        logger.info("Automatically turning spatial_subset to True based on inputs.")
        spatial_subset = True

    # If requested, call xscen.spatial.subset prior to averaging
    if spatial_subset:
        ds = subset(ds, **region)

    if method == "cos-lat":
        if ds.cf["latitude"].attrs.get("units", "") != "degrees_north":
            warnings.warn(
                f"Latitude units are '{ds.cf['latitude'].attrs.get('units', '')}', expected 'degrees_north'. "
                f"Make sure that the computation is right."
            )

        if ((ds.cf["longitude"].min() < -160) & (ds.cf["longitude"].max() > 160)) or (
            (ds.cf["longitude"].min() < 20) & (ds.cf["longitude"].max() > 340)
        ):
            warnings.warn(
                "The region appears to be crossing the -180/180° meridian. Bounds computation is currently bugged in cf_xarray. "
                "Make sure that the computation is right."
            )

        weights = np.cos(np.deg2rad(ds.cf["latitude"]))
        if ds.cf["longitude"].ndim == 1:
            dims = ds.cf["longitude"].dims + ds.cf["latitude"].dims
        else:
            if "longitude" not in ds.cf.bounds:
                ds = ds.cf.add_bounds(["longitude", "latitude"])
            # Weights the weights by the cell area (in °²)
            weights = weights * xr.DataArray(
                shapely.area(
                    shapely.polygons(shapely.linearrings(ds.lon_bounds, ds.lat_bounds))
                ),
                dims=ds.cf["longitude"].dims,
                coords=ds.cf["longitude"].coords,
            )
            dims = ds.cf["longitude"].dims
        ds_agg = ds.weighted(weights).mean(dims, keep_attrs=True)

        # Prepare the History field
        new_history = (
            f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
            f"weighted mean(dim={dims}) using a 'cos-lat' approximation of areacella (in deg2)"
        )

    elif method == "interp_centroid":
        # FIXME: No backward compatibility for this method since it produced incorrect results, but we want a specific error message for now.
        raise ValueError(
            "Method 'interp_centroid' has been removed, since it produced incorrect results."
        )

    # Uses xesmf.SpatialAverager
    elif method == "xesmf":
        if xe is None:
            raise ImportError(
                "Method xesmf requires xESMF to work, please install that package."
            )

        # If the region is a bounding box, call shapely and geopandas to transform it into an input compatible with xesmf
        if region["method"] == "bbox":
            polygon_geom = shapely.box(
                region["lon_bnds"][0],
                region["lat_bnds"][0],
                region["lon_bnds"][1],
                region["lat_bnds"][1],
            )
            polygon = gpd.GeoDataFrame(index=[0], geometry=[polygon_geom])

            # Prepare the History field
            new_history = (
                f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
                f"xesmf.SpatialAverager over {region['lon_bnds']}{region['lat_bnds']} - xESMF v{xe.__version__}"
            )

        # If the region is a shapefile, open with geopandas
        elif region["method"] == "shape":
            if not isinstance(region["shape"], gpd.GeoDataFrame):
                polygon = gpd.read_file(region["shape"])
                name = Path(region["shape"]).name
            else:
                polygon = region["shape"]
                name = f"{len(polygon)} polygons"

            # Simplify the geometries to a given tolerance, if needed.
            # The simpler the polygons, the faster the averaging, but it will lose some precision.
            if simplify_tolerance is not None:
                polygon = polygon.copy()
                polygon["geometry"] = polygon.simplify(
                    tolerance=simplify_tolerance, preserve_topology=True
                )

            # Prepare the History field
            new_history = (
                f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
                f"xesmf.SpatialAverager over {name} - xESMF v{xe.__version__}"
            )

        else:
            raise ValueError("'method' should be one of [bbox, shape].")

        kwargs_copy = deepcopy(kwargs)
        call_kwargs = {"skipna": kwargs_copy.pop("skipna", False)}
        if "output_chunks" in kwargs:
            call_kwargs["output_chunks"] = kwargs_copy.pop("output_chunks")

        # Preemptive segmentization. Same threshold as xESMF, but there isn't strong analysis behind this choice
        geoms = shapely.segmentize(polygon.geometry, 1)

        if (
            ds.cf["longitude"].ndim == 2
            and "longitude" not in ds.cf.bounds
            and "rotated_pole" in ds
        ):
            from .regrid import create_bounds_gridmapping

            ds = ds.update(create_bounds_gridmapping(ds))

        savg = xe.SpatialAverager(ds, geoms, **kwargs_copy)
        ds_agg = savg(ds, keep_attrs=True, **call_kwargs)
        extra_coords = {
            col: xr.DataArray(polygon[col], dims=("geom",))
            for col in polygon.columns
            if col != "geometry"
        }
        extra_coords["geom"] = xr.DataArray(polygon.index, dims=("geom",))
        ds_agg = ds_agg.assign_coords(**extra_coords)
        if len(polygon) == 1:
            ds_agg = ds_agg.squeeze("geom")
        if "lon_bounds" in ds_agg:
            ds_agg = ds_agg.assign_coords(
                {"lon_bounds": ds_agg.lon_bounds, "lat_bounds": ds_agg.lat_bounds}
            )

    else:
        raise ValueError(
            "Subsetting method should be ['cos-lat', 'interp_centroid', 'xesmf']"
        )

    # If the dataset had a projection, remove it
    grid_mapping = get_grid_mapping(ds)
    if grid_mapping:
        ds_agg = ds_agg.drop_vars(grid_mapping)
        for v in ds_agg.data_vars:
            ds_agg[v].attrs.pop("grid_mapping", None)

    # History
    history = (
        new_history + " \n " + ds_agg.attrs["history"]
        if "history" in ds_agg.attrs
        else new_history
    )
    ds_agg.attrs["history"] = history

    # Attrs
    if to_domain is not None:
        ds_agg.attrs["cat:domain"] = to_domain
    if to_level is not None:
        ds_agg.attrs["cat:processing_level"] = to_level

    return ds_agg


@parse_config
def produce_horizon(  # noqa: C901
    ds: xr.Dataset,
    *,
    indicators: (
        str
        | os.PathLike
        | Sequence[Indicator]
        | Sequence[tuple[str, Indicator]]
        | ModuleType
        | None
    ) = None,
    periods: list[str] | list[list[str]] | None = None,
    warminglevels: dict | None = None,
    op: str | dict = "mean",
    to_level: str | None = "horizons",
) -> xr.Dataset:
    """
    Compute indicators, then the climatological mean, and finally unstack dates in order
    to have a single dataset with all indicators of different frequencies.

    Once this is done, the function drops 'time' in favor of 'horizon'.
    This function computes the indicators and does an interannual mean.
    It stacks the season and month in different dimensions and adds a dimension `horizon` for the period or the warming level, if given.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset with a time dimension.
        If 'indicators' is None, the dataset should contain the precomputed indicators.
    indicators :  str | os.PathLike | Sequence[Indicator] | Sequence[Tuple[str, Indicator]] | ModuleType, optional
        Indicators to compute. It will be passed to the `indicators` argument of `xs.compute_indicators`.
    periods : list of str or list of lists of str, optional
        Either [start, end] or list of [start_year, end_year] for the period(s) to be evaluated.
        If both periods and warminglevels are None, the full time series will be used.
    warminglevels : dict, optional
        Dictionary of arguments to pass to `py:func:xscen.subset_warming_level`.
        If 'wl' is a list, the function will be called for each value and produce multiple horizons.
        If both periods and warminglevels are None, the full time series will be used.
    op : str or dict
        Operation to perform over the time dimension. See `py:func:xscen.climatological_op` for details. Default is 'mean'.
    to_level : str, optional
        The processing level to assign to the output.
        If there is only one horizon, you can use "{wl}", "{period0}" and "{period1}" in the string to dynamically
        include that information in the processing level.

    Returns
    -------
    xr.Dataset
        Horizon dataset.
    """
    if "warminglevel" in ds and len(ds.warminglevel) != 1:
        raise ValueError(
            "Input dataset should only have `warminglevel` dimension of length 1. "
            "If you want to use produce_horizon for multiple warming levels, "
            "extract the full time series and use the `warminglevels` argument instead."
        )

    all_periods = []
    if periods is not None:
        all_periods.extend(standardize_periods(periods))
    if warminglevels is not None:
        if isinstance(warminglevels["wl"], int | float):
            all_periods.append(warminglevels)
        elif isinstance(warminglevels["wl"], list):
            template = deepcopy(warminglevels)
            for wl in warminglevels["wl"]:
                all_periods.append({**template, "wl": wl})
        else:
            raise ValueError(
                f"Could not understand the format of warminglevels['wl']: {warminglevels['wl']}"
            )
    if len(all_periods) == 0:
        all_periods = standardize_periods(
            [[int(ds.time.dt.year[0]), int(ds.time.dt.year[-1])]]
        )

    out = []
    for period in all_periods:
        if isinstance(period, list):
            if ds.time.dt.year[0] <= int(period[0]) and ds.time.dt.year[-1] >= int(
                period[1]
            ):
                ds_sub = ds.sel(time=slice(period[0], period[1]))
            else:
                warnings.warn(
                    f"The requested period {period} is not fully covered by the input dataset. "
                    "The requested period will be skipped."
                )
                ds_sub = None
        else:
            ds_sub = subset_warming_level(ds, **period)

        if ds_sub is not None:
            if indicators is not None:
                # compute indicators
                ind_dict = compute_indicators(
                    ds=(
                        ds_sub.squeeze(dim="warminglevel")
                        if "warminglevel" in ds_sub.dims
                        else ds_sub
                    ),
                    indicators=indicators,
                )
            else:
                ind_dict = {"skipped": ds_sub}

            # Compute the window-year mean
            ds_merge = xr.Dataset()
            for freq, ds_ind in ind_dict.items():
                if freq != "fx":
                    ds_mean = climatological_op(
                        ds_ind,
                        op=op,
                        rename_variables=False,
                        horizons_as_dim=True,
                    ).drop_vars("time")
                else:
                    ds_mean = ds_ind.expand_dims(
                        dim={
                            "horizon": [
                                f"{ds.time.dt.year[0].item()}-{ds.time.dt.year[-1].item()}"
                            ]
                        }
                    )
                    ds_mean["horizon"] = ds_mean["horizon"].astype(str)

                if "warminglevel" in ds_mean.coords:
                    wl = np.array([ds_mean["warminglevel"].item()])
                    wl_attrs = ds_mean["warminglevel"].attrs
                    if "warminglevel" in ds_mean.dims:
                        ds_mean = ds_mean.squeeze("warminglevel")
                    ds_mean = ds_mean.drop_vars("warminglevel")
                    ds_mean["horizon"] = wl
                    ds_mean["horizon"].attrs.update(wl_attrs)

                # put all indicators in one dataset
                for var in ds_mean.data_vars:
                    ds_merge[var] = ds_mean[var]
                ds_merge.attrs.update(ds_mean.attrs)

            out.append(ds_merge)

    # If automatic attributes are not the same for all indicators, warn the user.
    if len(out) > 0:
        for v in out[0].data_vars:
            if not all(
                [
                    all(
                        [
                            out[0][v].attrs.get(attr) == out[i][v].attrs.get(attr)
                            for i in range(1, len(out))
                        ]
                    )
                    for attr in ["long_name", "description"]
                ]
            ):
                warnings.warn(
                    f"The attributes for variable {v} are not the same for all horizons, "
                    "probably because the periods were not of the same length. "
                    "Attributes will be kept from the first horizon, but this might not be the most appropriate."
                )

        out = xr.concat(out, dim="horizon")

        out.attrs["cat:xrfreq"] = "fx"
        out.attrs["cat:frequency"] = "fx"
        if to_level:
            if len(all_periods) == 1:
                if (isinstance(all_periods[0], dict)) or (
                    "warminglevel" in ds.dims and warminglevels is None
                ):
                    to_level = to_level.format(
                        wl=(
                            ds_sub.warminglevel.values[0]
                            if isinstance(all_periods[0], dict)
                            else ds.warminglevel.values[0]
                        )
                    )
                else:
                    to_level = to_level.format(
                        period0=all_periods[0][0], period1=all_periods[0][1]
                    )
            out.attrs["cat:processing_level"] = to_level

        return out

    else:
        raise ValueError("No horizon could be computed. Check your inputs.")
